"""High Level Orbit Interlock main application."""

import os as _os
import logging as _log
import time as _time
from functools import partial as _part

import numpy as _np

from ..util import update_bit as _updt_bit, get_bit as _get_bit
from ..namesys import SiriusPVName as _PVName
from ..search import LLTimeSearch as _LLTimeSearch, \
    HLTimeSearch as _HLTimeSearch
from ..thread import RepeaterThread as _Repeat, \
    LoopQueueThread as _LoopQueueThread
from ..epics import CAThread as _CAThread
from ..callbacks import Callback as _Callback
from ..devices import OrbitInterlock as _OrbitIntlk, FamBPMs as _FamBPMs, \
    EVG as _EVG, ASLLRF as _ASLLRF, Trigger as _Trigger, Device as _Device, \
    SOFB as _SOFB, HLFOFB as _FOFB, AFCPhysicalTrigger as _AFCPhysicalTrigger

from .csdev import Const as _Const, ETypes as _ETypes


class App(_Callback):
    """High Level Orbit Interlock main application."""

    SCAN_FREQUENCY = 1  # [Hz]

    def __init__(self, tests=False):
        """Class constructor."""
        super().__init__()
        self._is_dry_run = tests
        self._const = _Const()
        self._pvs_database = self._const.get_database()
        self._init = False

        # internal states
        self._llrf_intlk_state = 0b111011 if self._is_dry_run else 0b000000
        self._state = self._const.OffOn.Off
        self._bpm_status = self._pvs_database['BPMStatus-Mon']['value']
        self._timing_status = self._pvs_database['TimingStatus-Mon']['value']
        self._enable_lists = {
            'pos': _np.zeros(self._const.nr_bpms, dtype=bool),
            'ang': _np.zeros(self._const.nr_bpms, dtype=bool),
            'minsum': _np.zeros(self._const.nr_bpms, dtype=bool),
            }
        self._limits = {
            'pos_x_min': _np.zeros(self._const.nr_bpms, dtype=int),
            'pos_x_max': _np.zeros(self._const.nr_bpms, dtype=int),
            'pos_y_min': _np.zeros(self._const.nr_bpms, dtype=int),
            'pos_y_max': _np.zeros(self._const.nr_bpms, dtype=int),
            'ang_x_min': _np.zeros(self._const.nr_bpms, dtype=int),
            'ang_x_max': _np.zeros(self._const.nr_bpms, dtype=int),
            'ang_y_min': _np.zeros(self._const.nr_bpms, dtype=int),
            'ang_y_max': _np.zeros(self._const.nr_bpms, dtype=int),
            'minsum': _np.zeros(self._const.nr_bpms, dtype=int),
            }
        self._acq_chan = self._pvs_database['PsMtmAcqChannel-Sel']['value']
        self._acq_spre = self._pvs_database['PsMtmAcqSamplesPre-SP']['value']
        self._acq_spost = self._pvs_database['PsMtmAcqSamplesPost-SP']['value']
        self._thread_acq = None
        self._thread_cbevgilk = None
        self._thread_cbevgrx = None
        self._thread_cbbpm = None
        self._bpm_mon_devs = list()
        self._ti_mon_devs = list()
        self._lock_threads = dict()
        self._lock_failures = set()
        self._lock_suspend = True
        self._set_queue = _LoopQueueThread()
        self._set_queue.start()

        # devices and connections
        # # EVG
        self._evg_dev = _EVG(props2init=[
            'IntlkCtrlEnbl-Sel', 'IntlkCtrlEnbl-Sts',
            'IntlkCtrlRst-Sel', 'IntlkCtrlRst-Sts',
            'IntlkCtrlRepeat-Sel', 'IntlkCtrlRepeat-Sts',
            'IntlkCtrlRepeatTime-SP', 'IntlkCtrlRepeatTime-RB',
            'IntlkTbl0to15-Sel', 'IntlkTbl0to15-Sts',
            'IntlkTbl16to27-Sel', 'IntlkTbl16to27-Sts',
            'IntlkEvtIn0-SP', 'IntlkEvtIn0-RB',
            'IntlkEvtOut-SP', 'IntlkEvtOut-SP',
            'IntlkEvtStatus-Mon',
            'RxEnbl-SP', 'RxEnbl-RB',
            'RxEnbl-SP.B0', 'RxEnbl-RB.B0',
            'RxEnbl-SP.B1', 'RxEnbl-RB.B1',
            'RxEnbl-SP.B2', 'RxEnbl-RB.B2',
            'RxEnbl-SP.B3', 'RxEnbl-RB.B3',
            'RxEnbl-SP.B4', 'RxEnbl-RB.B4',
            'RxEnbl-SP.B5', 'RxEnbl-RB.B5',
            'RxEnbl-SP.B6', 'RxEnbl-RB.B6',
            'RxEnbl-SP.B7', 'RxEnbl-RB.B7',
            'RxLockedLtc-Mon', 'RxLockedLtcRst-Cmd',
            ])
        # interlock callback
        pvo = self._evg_dev.pv_object('IntlkEvtStatus-Mon')
        pvo.auto_monitor = True
        pvo.add_callback(self._callback_evg_intlk)
        pvo.connection_callbacks.append(self._conn_callback_timing)
        # rxlock callback
        pvo = self._evg_dev.pv_object('RxLockedLtc-Mon')
        pvo.auto_monitor = True
        pvo.add_callback(self._callback_evg_rxlock)

        # # Fouts
        foutnames = list(self._const.FOUTS_2_MON) + \
            list(self._const.FOUTSFIXED_RXENBL.keys())
        self._thread_cbfout = {fout: None for fout in foutnames}
        self._fout_devs = {
            devname: _Device(
                devname,
                props2init=[
                    'RxEnbl-SP', 'RxEnbl-RB',
                    'RxLockedLtc-Mon', 'RxLockedLtcRst-Cmd',
                ], auto_monitor_mon=True)
            for devname in foutnames}
        self._fout2rxenbl = dict()
        for devname, dev in self._fout_devs.items():
            pvo = dev.pv_object('RxLockedLtc-Mon')
            pvo.add_callback(self._callback_fout_rxlock)
            pvo.connection_callbacks.append(self._conn_callback_timing)
            rxenbl = self._const.FOUTSFIXED_RXENBL.get(devname, 0)
            self._fout2rxenbl[devname] = rxenbl

        # # AFC timing
        self._afcti_devs = {
            idx+1: _Device(
                f'IA-{idx+1:02}RaBPM:TI-AMCFPGAEVR',
                props2init=[
                    'DevEnbl-Sel', 'DevEnbl-Sts',
                    'RTMClkLockedLtc-Mon', 'ClkLockedLtcRst-Cmd',
                    'RTMClkRst-Cmd',
                    'RTMPhasePropGain-SP', 'RTMPhasePropGain-RB',
                    'RTMPhaseIntgGain-SP', 'RTMPhaseIntgGain-RB',
                    'RTMFreqPropGain-SP', 'RTMFreqPropGain-RB',
                    'RTMFreqIntgGain-SP', 'RTMFreqIntgGain-RB',
                    'RTMPhaseNavg-SP', 'RTMPhaseNavg-RB',
                    'RTMPhaseDiv-SP', 'RTMPhaseDiv-RB',
                ], auto_monitor_mon=True)
            for idx in range(20)
        }
        for dev in self._afcti_devs.values():
            pvo = dev.pv_object('RTMClkLockedLtc-Mon')
            pvo.connection_callbacks.append(self._conn_callback_timing)

        # # RF EVE
        trgsrc = _HLTimeSearch.get_ll_trigger_names('SI-Glob:TI-LLRF-PsMtm')
        pvname = _LLTimeSearch.get_channel_output_port_pvname(trgsrc[0])
        self._llrf_evtcnt_pvname = f'{pvname.propty}EvtCnt-Mon'
        self._everf_dev = _Device(
            pvname.device_name,
            props2init=[self._llrf_evtcnt_pvname, ],
            auto_monitor_mon=True)
        pvo = self._everf_dev.pv_object(self._llrf_evtcnt_pvname)
        pvo.wait_for_connection()
        self._everf_evtcnt = pvo.get() or 0

        # # HL triggers
        self._hltrig_devs = dict()
        for trigname, configs in self._const.HLTRIG_2_CONFIG:
            props2init = list()
            for prop, _ in configs:
                props2init.append(prop)
                props2init.append(_PVName.from_sp2rb(prop))
            props2init.append('Status-Mon')
            self._hltrig_devs[trigname] = _Trigger(
                trigname=trigname,
                props2init=props2init,
                auto_monitor_mon=True)
            if 'LLRF' in trigname or 'OrbIntlkRedundancy' in trigname:
                pvo = self._hltrig_devs[trigname].pv_object('Status-Mon')
                pvo.add_callback(self._callback_hltrig_status)

        # # BPM devices
        self._orbintlk_dev = _OrbitIntlk()
        for dev in self._orbintlk_dev.devices:
            pvo = dev.pv_object('IntlkLtc-Mon')
            pvo.auto_monitor = True
            pvo.add_callback(self._callback_bpm_intlk)

        self._fambpm_dev = _FamBPMs(
            devname=_FamBPMs.DEVICES.SI, ispost_mortem=True,
            props2init=[
                'ACQChannel-Sel', 'ACQChannel-Sts',
                'ACQSamplesPre-SP', 'ACQSamplesPre-RB',
                'ACQSamplesPost-SP', 'ACQSamplesPost-RB',
                'ACQTriggerRep-Sel', 'ACQTriggerRep-Sts',
                'ACQTrigger-Sel', 'ACQTrigger-Sts',
                'ACQTriggerEvent-Sel', 'ACQTriggerEvent-Sts',
                'ACQStatus-Sts',
                'INFOFAcqRate-RB', 'INFOMONITRate-RB',
                'TRIGGER4TrnSrc-Sel', 'TRIGGER4TrnSrc-Sts',
                'TRIGGER4TrnOutSel-SP', 'TRIGGER4TrnOutSel-RB',
                'TRIGGER_PM0RcvSrc-Sel', 'TRIGGER_PM0RcvSrc-Sts',
                'TRIGGER_PM0RcvInSel-SP', 'TRIGGER_PM0RcvInSel-RB',
                'TRIGGER_PM1RcvSrc-Sel', 'TRIGGER_PM1RcvSrc-Sts',
                'TRIGGER_PM1RcvInSel-SP', 'TRIGGER_PM1RcvInSel-RB',
                'TRIGGER_PM6RcvSrc-Sel', 'TRIGGER_PM6RcvSrc-Sts',
                'TRIGGER_PM6RcvInSel-SP', 'TRIGGER_PM6RcvInSel-RB',
                'TRIGGER_PM7RcvSrc-Sel', 'TRIGGER_PM7RcvSrc-Sts',
                'TRIGGER_PM7RcvInSel-SP', 'TRIGGER_PM7RcvInSel-RB',
                'TRIGGER_PM11RcvSrc-Sel', 'TRIGGER_PM11RcvSrc-Sts',
                'TRIGGER_PM11RcvInSel-SP', 'TRIGGER_PM11RcvInSel-RB',
                'TRIGGER_PM12RcvSrc-Sel', 'TRIGGER_PM12RcvSrc-Sts',
                'TRIGGER_PM12RcvInSel-SP', 'TRIGGER_PM12RcvInSel-RB',
                'TRIGGER_PM14RcvSrc-Sel', 'TRIGGER_PM14RcvSrc-Sts',
                'TRIGGER_PM14RcvInSel-SP', 'TRIGGER_PM14RcvInSel-RB',
                'ADCAD9510PllStatus-Mon'])
        self._monitsum2intlksum_factor = 0
        for dev in self._fambpm_dev.devices:
            pvo = dev.pv_object('ADCAD9510PllStatus-Mon')
            pvo.auto_monitor = True
            pvo.add_callback(self._callback_bpm_adclock)

        # # AFC physical trigger devices
        phytrig_names = list()
        for afcti, cratemap in self._const.crates_map.items():
            if '20RaBPMTL' in afcti:
                continue
            phytrig_names.extend(cratemap)
        self._phytrig_devs = [
            _AFCPhysicalTrigger(dev, 4, props2init=[
                'Dir-Sel', 'Dir-Sts',
                'DirPol-Sel', 'DirPol-Sts',
                'TrnLen-SP', 'TrnLen-RB'])
            for dev in phytrig_names]
        for dev in self._phytrig_devs:
            pvo = dev.pv_object('Dir-Sel')
            pvo.connection_callbacks.append(self._conn_callback_afcphystrigs)

        # # RF devices
        self._llrf = _ASLLRF(
            devname=_ASLLRF.DEVICES.SI,
            props2init=[
                'ILK:BEAM:TRIP:S', 'ILK:BEAM:TRIP', 'FASTINLK-MON',
                'ILK:MAN:S', 'ILK:MAN', 'IntlkSet-Cmd', 'Reset-Cmd',
            ])
        self._llrf.pv_object('FASTINLK-MON').auto_monitor = True

        # # auxiliary devices
        self._fofb = _FOFB(
            props2init=['LoopState-Sts', ])
        self._sofb = _SOFB(
            _SOFB.DEVICES.SI,
            props2init=['LoopState-Sts', 'SlowSumRaw-Mon'])
        self._sofb.pv_object('SlowSumRaw-Mon').auto_monitor = True

        # pvs to write methods
        self.map_pv2write = {
            'Enable-Sel': self.set_enable,
            'PosEnblList-SP': _part(self.set_enbllist, 'pos'),
            'AngEnblList-SP': _part(self.set_enbllist, 'ang'),
            'MinSumEnblList-SP': _part(self.set_enbllist, 'minsum'),
            'PosXMinLim-SP': _part(self.set_intlk_lims, 'pos_x_min'),
            'PosXMaxLim-SP': _part(self.set_intlk_lims, 'pos_x_max'),
            'PosYMinLim-SP': _part(self.set_intlk_lims, 'pos_y_min'),
            'PosYMaxLim-SP': _part(self.set_intlk_lims, 'pos_y_max'),
            'AngXMinLim-SP': _part(self.set_intlk_lims, 'ang_x_min'),
            'AngXMaxLim-SP': _part(self.set_intlk_lims, 'ang_x_max'),
            'AngYMinLim-SP': _part(self.set_intlk_lims, 'ang_y_min'),
            'AngYMaxLim-SP': _part(self.set_intlk_lims, 'ang_y_max'),
            'MinSumLim-SP': _part(self.set_intlk_lims, 'minsum'),
            'ResetBPMGen-Cmd': _part(self.cmd_reset, 'bpm_gen'),
            'ResetBPMPos-Cmd': _part(self.cmd_reset, 'bpm_pos'),
            'ResetBPMAng-Cmd': _part(self.cmd_reset, 'bpm_ang'),
            'ResetBPM-Cmd': _part(self.cmd_reset, 'bpm_all'),
            'Reset-Cmd': _part(self.cmd_reset, 'all'),
            'PsMtmAcqChannel-Sel': self.set_acq_channel,
            'PsMtmAcqSamplesPre-SP': self.set_acq_nrspls_pre,
            'PsMtmAcqSamplesPost-SP': self.set_acq_nrspls_post,
            'PsMtmAcqConfig-Cmd': self.cmd_acq_config,
            'ResetTimingLockLatches-Cmd': self.cmd_reset_ti_lock_latch,
            'ResetAFCTimingRTMClk-Cmd': self.cmd_reset_afcti_rtmclk,
            'ConfigEVG-Cmd': self.cmd_config_evg,
            'ConfigFouts-Cmd': self.cmd_config_fouts,
            'ConfigAFCTiming-Cmd': self.cmd_config_afcti,
            'ConfigHLTriggers-Cmd': self.cmd_config_hltrigs,
            'ConfigLLRFIntlk-Cmd': self.cmd_config_llrf,
            'ConfigBPMs-Cmd': self.cmd_config_bpms,
            'ConfigAFCPhyTrigs-Cmd': self.cmd_config_phytrigs,
            }

        # configuration scanning
        self.thread_check_configs = _Repeat(
            1.0/App.SCAN_FREQUENCY, self._check_configs, niter=0,
            is_cathread=True)
        self.thread_check_configs.pause()
        self.thread_check_configs.start()

    def init_database(self):
        """Set initial PV values."""
        pvn2vals = {
            'Enable-Sel': self._state,
            'Enable-Sts': self._state,
            'BPMStatus-Mon': self._bpm_status,
            'TimingStatus-Mon': self._timing_status,
            'ResetBPMGen-Cmd': 0,
            'ResetBPMPos-Cmd': 0,
            'ResetBPMAng-Cmd': 0,
            'ResetBPM-Cmd': 0,
            'Reset-Cmd': 0,
            'PsMtmAcqChannel-Sel': self._acq_chan,
            'PsMtmAcqChannel-Sts': self._acq_chan,
            'PsMtmAcqSamplesPre-SP': self._acq_spre,
            'PsMtmAcqSamplesPre-RB': self._acq_spre,
            'PsMtmAcqSamplesPost-SP': self._acq_spost,
            'PsMtmAcqSamplesPost-RB': self._acq_spost,
            'PsMtmAcqConfig-Cmd': 0,
        }
        for pvn, val in pvn2vals.items():
            self.run_callbacks(pvn, val)

        # load autosave data

        # enable lists
        for ilk in ['Pos', 'Ang', 'MinSum']:
            ilkname = ilk.lower()
            okl = self._load_file(ilkname, 'enbl')
            pvn = f'{ilk}EnblList'
            enb = self._enable_lists[ilkname]
            self.run_callbacks(pvn+'-SP', enb)
            if not okl:
                self.run_callbacks(pvn+'-RB', enb)
        self._bpm_mon_devs, self._ti_mon_devs = self._get_monitored_devices()
        self.run_callbacks(
            'BPMMonitoredDevices-Mon', '\n'.join(self._bpm_mon_devs))
        self.run_callbacks(
            'TimingMonitoredDevices-Mon', '\n'.join(self._ti_mon_devs))
        self._config_fout_rxenbl()

        # limits
        for ilk in ['Pos', 'Ang']:
            for pln in ['X', 'Y']:
                for lim in ['Min', 'Max']:
                    atn = f'{ilk}_{pln}_{lim}'.lower()
                    pvn = f'{ilk}{pln}{lim}Lim'
                    okl = self._load_file(atn, 'lim')
                    val = self._limits[atn]
                    self.run_callbacks(pvn+'-SP', val)
                    if not okl:
                        self.run_callbacks(pvn+'-RB', val)

        okl = self._load_file('minsum', 'lim')
        val = self._limits['minsum']
        self.run_callbacks('MinSumLim-SP', val)
        if not okl:
            self.run_callbacks('MinSumLim-RB', val)

        # wait while enable list and limits setpoint queue to be empty
        self._set_queue.join()

        self._update_log('Started.')
        self._init = True

        # start init lock devices
        self._enable_lock(init=True)

    def _enable_lock(self, init=False):
        if not init:
            self._lock_suspend = False
        self._handle_lock_evg_configs(init)
        self._handle_lock_fouts(init)
        self._handle_lock_afcti(init)
        self._handle_lock_hltriggers(init)
        self._handle_lock_llrf(init)
        self._handle_lock_bpm_configs(init)
        self._handle_lock_afcphytrigs(init)
        if init:
            self._handle_lock_evg_enable(init)
            self._handle_lock_bpm_enable(init)

    def _disable_lock(self):
        self._lock_suspend = True

    def _handle_lock_evg_configs(self, init=False):
        self._evg_dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
        for propty_sp, desired_val in self._const.EVG_CONFIGS:
            propty_rb = _PVName.from_sp2rb(propty_sp)
            pvo = self._evg_dev.pv_object(propty_rb)
            if init:
                pvo.add_callback(_part(
                    self._callback_lock, self._evg_dev, propty_sp, desired_val))
            else:
                pvo.run_callbacks()

    def _handle_lock_evg_enable(self, init=False):
        # lock interlock enable state
        pvo = self._evg_dev.pv_object('IntlkCtrlEnbl-Sts')
        if init:
            pvo.add_callback(self._callback_evg_lock_intlk)
        else:
            pvo.run_callbacks()

    def _handle_lock_fouts(self, init=False):
        for dev in self._fout_devs.values():
            dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
            pvo = dev.pv_object('RxEnbl-RB')
            if init:
                pvo.add_callback(self._callback_fout_lock)
            else:
                pvo.run_callbacks()

    def _handle_lock_afcti(self, init=False):
        for dev in self._afcti_devs.values():
            dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
            for propty_sp, desired_val in self._const.AFCTI_CONFIGS:
                propty_rb = _PVName.from_sp2rb(propty_sp)
                pvo = dev.pv_object(propty_rb)
                if init:
                    pvo.add_callback(_part(
                        self._callback_lock, dev, propty_sp, desired_val))
                else:
                    pvo.run_callbacks()

    def _handle_lock_hltriggers(self, init=False):
        for trigname, configs in self._const.HLTRIG_2_CONFIG:
            trigdev = self._hltrig_devs[trigname]
            trigdev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
            for prop_sp, desired_val in configs:
                prop_rb = _PVName.from_sp2rb(prop_sp)
                pvo = trigdev.pv_object(prop_rb)
                if init:
                    pvo.add_callback(
                        _part(self._callback_lock, trigdev, prop_sp, desired_val))
                else:
                    pvo.run_callbacks()

    def _handle_lock_llrf(self, init=False):
        self._llrf.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
        pvo_beamtrip = self._llrf.pv_object('ILK:BEAM:TRIP')
        pvo_manintlk = self._llrf.pv_object('ILK:MAN')
        if init:
            pvo_beamtrip.add_callback(_part(
                self._callback_lock, self._llrf,
                'ILK:BEAM:TRIP:S', self._llrf_intlk_state))
            pvo_manintlk.add_callback(_part(
                self._callback_lock, self._llrf,
                'ILK:MAN:S', self._llrf_intlk_state))
        else:
            pvo_beamtrip.run_callbacks()
            pvo_manintlk.run_callbacks()

    def _handle_lock_bpm_configs(self, init=False):
        # lock BPM interlock enable and limits
        prop2lock = [
            'IntlkMinSumEn-Sts',
            'IntlkLmtMinSum-RB',
            'IntlkPosEn-Sts',
            'IntlkLmtPosMaxX-RB',
            'IntlkLmtPosMinX-RB',
            'IntlkLmtPosMaxY-RB',
            'IntlkLmtPosMinY-RB',
            'IntlkAngEn-Sts',
            'IntlkLmtAngMaxX-RB',
            'IntlkLmtAngMinX-RB',
            'IntlkLmtAngMaxY-RB',
            'IntlkLmtAngMinY-RB',
        ]
        self._orbintlk_dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
        for dev in self._orbintlk_dev.devices:
            for prop in prop2lock:
                pvo = dev.pv_object(prop)
                if init:
                    pvo.add_callback(self._callback_bpm_lock)
                else:
                    pvo.run_callbacks()

        # lock BPM logical triggers
        self._fambpm_dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
        for dev in self._fambpm_dev.devices:
            for prop_sp, desired_val in self._const.SIBPMLOGTRIG_CONFIGS:
                prop_rb = _PVName.from_sp2rb(prop_sp)
                pvo = dev.pv_object(prop_rb)
                if init:
                    pvo.add_callback(
                        _part(self._callback_lock, dev, prop_sp, desired_val))
                else:
                    pvo.run_callbacks()

    def _handle_lock_bpm_enable(self, init=False):
        self._orbintlk_dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
        for dev in self._orbintlk_dev.devices:
            pvo = dev.pv_object('IntlkEn-Sts')
            if init:
                pvo.add_callback(self._callback_bpm_lock)
            else:
                pvo.run_callbacks()

    def _handle_lock_afcphytrigs(self, init=False):
        for dev in self._phytrig_devs:
            dev.wait_for_connection(timeout=self._const.DEF_TIMEOUT)
            for prop_sp, desired_val in self._const.AFCPHYTRIG_CONFIGS:
                # only lock polarity of other AFC physical triggers than SI BPM
                if not self._check_lock_phytrig_prop(dev, prop_sp):
                    continue
                prop_rb = _PVName.from_sp2rb(prop_sp)
                pvo = dev.pv_object(prop_rb)
                if init:
                    pvo.add_callback(
                        _part(self._callback_lock, dev, prop_sp, desired_val))
                else:
                    pvo.run_callbacks()

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        _ = reason
        return None

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        _log.info('Write received for: %s --> %s', reason, str(value))
        if reason not in self.map_pv2write:
            _log.warning('PV %s does not have a set function.', reason)
            return False

        status = self.map_pv2write[reason](value)
        _log.info(
            '%s Write for: %s --> %s', str(status).upper(), reason, str(value))
        return status

    @property
    def evg_dev(self):
        """EVG device."""
        return self._evg_dev

    @property
    def orbintlk_dev(self):
        """Orbit interlock device."""
        return self._orbintlk_dev

    @property
    def fambpm_dev(self):
        """Return FamBPMs device."""
        return self._fambpm_dev

    # --- interlock control ---

    def set_enable(self, value):
        """Set orbit interlock state.
        Configure global BPM interlock enable and EVG interlock enable."""
        self._set_queue.put((self._do_set_enable, (value, )))
        return True

    def _do_set_enable(self, value):
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            self.run_callbacks('Enable-Sel', self._state)
            return False

        if value:
            if not self._check_ti_devices_status(self._ti_mon_devs):
                self._update_log('ERR:Could not enable orbit interlock.')
                self.run_callbacks('Enable-Sel', self._state)
                return False
            glob_en = self._get_gen_bpm_intlk()
        else:
            glob_en = _np.zeros(self._const.nr_bpms, dtype=bool)

        bkup = int(self._state)
        self._state = value

        if self._state:
            self._enable_lock()
        else:
            self._disable_lock()

        if not self._orbintlk_dev.set_gen_enable(list(glob_en)):
            self._update_log('ERR:Could not set BPM general')
            self._update_log('ERR:interlock enable.')
            self._state = bkup
            self.run_callbacks('Enable-Sel', self._state)
            return False
        self._update_log('Configured BPM general interlock enable.')

        self._evg_dev['IntlkCtrlEnbl-Sel'] = value
        if not self._evg_dev._wait(
                'IntlkCtrlEnbl-Sts', value, timeout=self._const.DEF_TIMEOUT):
            self._update_log('ERR:Could not set EVG interlock enable.')
            self._state = bkup
            self.run_callbacks('Enable-Sel', self._state)
            return False
        self._update_log('Configured EVG interlock enable.')

        self.run_callbacks('Enable-Sts', self._state)

        return True

    # --- enable lists ---

    def set_enbllist(self, intlk, value):
        """Set enable list for interlock type."""
        if self._state:
            self._update_log('ERR:Disable interlock before changing')
            self._update_log('ERR:enable lists.')
            return False
        self._set_queue.put((self._do_set_enbllist, (intlk, value)))
        return True

    def _do_set_enbllist(self, intlk, value):
        intlkname = intlk.capitalize().replace('sum', 'Sum')
        self._update_log(f'Setting {intlkname} EnblList...')

        # check size
        new = _np.array(value, dtype=bool)
        if self._const.nr_bpms != new.size:
            self._update_log(f'ERR:Wrong {intlkname} EnblList size.')
            self.run_callbacks(
                f'{intlkname}EnblList-SP', self._enable_lists[intlk])
            return False

        # check coerence, down/up pair should have same enable state
        if not self._check_valid_bpmconfig(new):
            self._update_log('ERR:BPM should be enabled in pairs')
            self._update_log('ERR:(M1/M2,C1-1/C1-2,C2/C3-1,C3-2/C4)')
            self.run_callbacks(
                f'{intlkname}EnblList-SP', self._enable_lists[intlk])
            return False

        bkup_enbllist = self._enable_lists[intlk]
        self._enable_lists[intlk] = new

        # do not write to devices and save to file in initialization
        if not self._init:
            self._update_log('...done.')
            # update readback pv
            self.run_callbacks(f'{intlkname}EnblList-RB', new)
            return True

        # check if new enable list do not imply in orbit interlock failure
        if intlk in ['pos', 'ang']:
            bkup_bpmmon, bkup_timon = self._bpm_mon_devs, self._ti_mon_devs
            self._bpm_mon_devs, self._ti_mon_devs = \
                self._get_monitored_devices()
            self._config_fout_rxenbl()
            if not self._check_ti_devices_status(self._ti_mon_devs):
                self._update_log('ERR:Could not set enable list.')
                self._enable_lists[intlk] = bkup_enbllist
                self._bpm_mon_devs, self._ti_mon_devs = bkup_bpmmon, bkup_timon
                self._config_fout_rxenbl()
                self.run_callbacks(f'{intlkname}EnblList-SP', bkup_enbllist)
                return False
            self.run_callbacks(
                'BPMMonitoredDevices-Mon', '\n'.join(self._bpm_mon_devs))
            self.run_callbacks(
                'TimingMonitoredDevices-Mon', '\n'.join(self._ti_mon_devs))

        # handle device enable configuration

        # set BPM interlock specific enable state
        fun = getattr(self._orbintlk_dev, f'set_{intlk}_enable')
        ret = fun(list(value), timeout=3, return_prob=True)
        if not ret[0]:
            self._update_log(f'ERR:Could not set BPM {intlkname}')
            self._update_log('ERR:interlock enable.')
            for item in ret[1]:
                self._update_log(f'ERR:Verify:{item}')
            self.run_callbacks(f'{intlkname}EnblList-SP', bkup_enbllist)
            return False

        # if interlock is already enabled, update BPM general enable state
        if self._state and intlk in ['pos', 'ang']:
            glob_en = self._get_gen_bpm_intlk()
            ret = self._orbintlk_dev.set_gen_enable(
                list(glob_en), timeout=3, return_prob=True)
            if not ret[0]:
                self._update_log('ERR:Could not set BPM general')
                self._update_log('ERR:interlock enable.')
                for item in ret[1]:
                    self._update_log(f'ERR:Verify:{item}')
                self.run_callbacks(f'{intlkname}EnblList-SP', bkup_enbllist)
                return False

        # save to autosave files
        self._save_file(intlk, _np.array([value], dtype=bool), 'enbl')

        self._update_log('...done.')

        # update readback pv
        self.run_callbacks(f'{intlkname}EnblList-RB', new)
        return True

    # --- limits ---

    def set_intlk_lims(self, intlk_lim, value):
        """Set limits for interlock type."""
        if self._state:
            self._update_log('ERR:Disable interlock before changing')
            self._update_log('ERR:interlock thresholds.')
            return False
        self._set_queue.put((self._do_set_intlk_lims, (intlk_lim, value)))
        return True

    def _do_set_intlk_lims(self, intlk_lim, value):
        parts = intlk_lim.split('_')
        if len(parts) > 1:
            ilk, pln, lim = parts
            limname = f'{ilk.capitalize()}{pln.capitalize()}{lim.capitalize()}'
        else:
            limname = intlk_lim.capitalize().replace('sum', 'Sum')
        self._update_log(f'Setting {limname} limits...')

        # check size
        new = _np.array(value, dtype=int)
        if self._const.nr_bpms != new.size:
            self._update_log(f'ERR: Wrong {limname} limits size.')
            self.run_callbacks(f'{limname}Lim-SP', self._limits[intlk_lim])
            return False

        # check coerence, down/up pair should have same limits
        if not self._check_valid_bpmconfig(new):
            self._update_log('ERR:BPM pairs should have equal limits')
            self._update_log('ERR:(M1/M2,C1-1/C1-2,C2/C3-1,C3-2/C4)')
            self.run_callbacks(f'{limname}Lim-SP', self._limits[intlk_lim])
            return False

        self._limits[intlk_lim] = new

        # do not set limits and save to file in initialization
        if not self._init:
            self._update_log('...done.')
            # update readback pv
            self.run_callbacks(f'{limname}Lim-RB', new)
            return True

        # handle device limits configuration

        # set BPM interlock limits
        fun = getattr(self._orbintlk_dev, f'set_{intlk_lim}_thres')
        ret = fun(list(value), timeout=3, return_prob=True)
        if not ret[0]:
            self._update_log(f'ERR:Could not set BPM {limname}')
            self._update_log('ERR:interlock limits.')
            for item in ret[1]:
                self._update_log(f'ERR:Verify:{item}')
            self.run_callbacks(f'{limname}Lim-SP', self._limits[intlk_lim])
            return False

        # save to autosave files
        self._save_file(intlk_lim, _np.array([value]), 'lim')

        self._update_log('...done.')

        # update readback pv
        self.run_callbacks(f'{limname}Lim-RB', new)
        return True

    # --- reset ---

    def cmd_reset(self, state, value=None):
        """Reset interlock states."""
        _ = value
        # if it is a BPM position, BPM general or a global reset
        if 'pos' in state or 'all' in state:
            self._orbintlk_dev.cmd_reset_pos()
            self._update_log('Sent reset BPM position flags.')
        # if it is a BPM angle, BPM general or a global reset
        if 'ang' in state or 'all' in state:
            self._orbintlk_dev.cmd_reset_ang()
            self._update_log('Sent reset BPM angle flags.')
        # if it is a BPM general or a global reset
        if 'gen' in state or 'all' in state:
            self._orbintlk_dev.cmd_reset_gen()
            self._update_log('Sent reset BPM general flags.')

        # if it is a global reset, reset EVG
        if state == 'all':
            self._evg_dev['IntlkCtrlRst-Sel'] = 1
            self._update_log('Sent reset EVG interlock flag.')

        return True

    def cmd_reset_ti_lock_latch(self, value=None):
        """Command to reset AFC timing and Fout clock lock latches."""
        _ = value
        # try to reset AFC timing clock lock latches, act only in necessary
        # devices, return false if fail
        for idx, afcti in self._afcti_devs.items():
            if afcti['RTMClkLockedLtc-Mon']:
                continue
            afcti['ClkLockedLtcRst-Cmd'] = 1
            msg = 'Reset' if afcti._wait('RTMClkLockedLtc-Mon', 1, timeout=3) \
                else 'ERR:Could not reset'
            self._update_log(f'{msg} AFC Timing {idx} lock latchs.')
            if 'not' in msg:
                return False
        # try to reset BPM Fout rx lock latches, act only in necessary
        # devices, return false if fail
        for devname, fout in self._fout_devs.items():
            rxv = self._fout2rxenbl[devname]
            if fout['RxLockedLtc-Mon'] == rxv:
                continue
            fout['RxLockedLtcRst-Cmd'] = 1
            msg = 'Reset' if fout._wait('RxLockedLtc-Mon', rxv, timeout=3) \
                else 'ERR:Could not reset'
            self._update_log(f'{msg} {devname} lock latchs.')
            if 'not' in msg:
                return False
        return True

    def cmd_reset_afcti_rtmclk(self, value=None):
        """Command to reset AFC timing clocks."""
        _ = value
        #  do not allow user to reset in case of correction loops closed
        if not self._fofb.connected or self._fofb['LoopState-Sts'] or \
                not self._sofb.connected or self._sofb['LoopState-Sts']:
            self._update_log('ERR:Open correction loops before ')
            self._update_log('ERR:reseting AFC Timing clocks.')
            return False

        # try to reset AFC timing clock, act only in necessary
        # devices, return false if fail
        for idx, afcti in self._afcti_devs.items():
            if afcti['RTMClkLockedLtc-Mon']:
                continue
            afcti['ClkLockedLtcRst-Cmd'] = 1
            if afcti._wait('RTMClkLockedLtc-Mon', 1, timeout=3):
                continue
            afcti['RTMClkRst-Cmd'] = 1
            self._update_log(f'Sent reset clock to AFC Timing {idx}.')

        # try to reset latches
        self.cmd_reset_ti_lock_latch()

        return True

    # --- BPM acquisition ---

    def set_acq_channel(self, value):
        """Set BPM PsMtm acquisition channel."""
        self._acq_chan = value
        self.run_callbacks('PsMtmAcqChannel-Sts', value)
        return True

    def set_acq_nrspls_pre(self, value):
        """Set BPM PsMtm acquisition number of samples pre."""
        self._acq_spre = value
        self.run_callbacks('PsMtmAcqSamplesPre-RB', value)
        return True

    def set_acq_nrspls_post(self, value):
        """Set BPM PsMtm acquisition number of samples post."""
        self._acq_spost = value
        self.run_callbacks('PsMtmAcqSamplesPost-RB', value)
        return True

    def cmd_acq_config(self, value=None):
        """Configure BPM PsMtm acquisition."""
        if self._thread_acq and self._thread_acq.is_alive():
            self._update_log('WARN:BPM configuration already in progress.')
            return False
        self._thread_acq = _CAThread(target=self._acq_config, daemon=True)
        self._thread_acq.start()
        return True

    def _acq_config(self):
        self._update_log('Aborting BPM acquisition...')
        ret = self._fambpm_dev.cmd_mturn_acq_abort()
        if ret > 0:
            self._update_log('ERR:Failed to abort BPM acquisition.')
            return
        self._update_log('...done. Configuring BPM acquisition...')
        ret = self._fambpm_dev.mturn_config_acquisition(
            nr_points_before=self._acq_spre,
            nr_points_after=self._acq_spost,
            acq_rate=self._acq_chan,
            repeat=False,
            external=True)
        if ret < 0:
            self._update_log(
                'ERR:Failed to abort acquisition for ' +
                f'{self._const.bpm_names[-ret-1]:s}.')
            return
        if ret > 0:
            self._update_log(
                'ERR:Failed to start acquisition for ' +
                f'{self._const.bpm_names[ret-1]:s}.')
            return
        self._update_log('...done!')

    # --- devices configurations ---

    def cmd_config_evg(self, value):
        """Configure EVG according to lock configurations."""
        _ = value
        if not self._evg_dev.connected:
            self._update_log('ERR:EVG disconnected.')
            return False
        for propty_sp, desired_val in self._const.EVG_CONFIGS:
            propty_rb = _PVName.from_sp2rb(propty_sp)
            if self._evg_dev[propty_rb] == desired_val:
                continue
            pvo = self._evg_dev.pv_object(propty_sp)
            pvo.put(desired_val, wait=True)
            _time.sleep(0.2)
        return True

    def cmd_config_fouts(self, value):
        """Configure Fouts according to lock configurations."""
        _ = value
        for devname, dev in self._fout_devs.items():
            if not dev.connected:
                self._update_log(f'ERR:{devname} disconnected.')
                continue
            desired_value = self._fout2rxenbl[devname]
            dev['RxEnbl-SP'] = desired_value
            dev._wait('RxEnbl-RB', desired_value, timeout=1)
            dev['RxLockedLtcRst-Cmd'] = 1
        return True

    def cmd_config_afcti(self, value):
        """Configure all AFC timing according to lock configurations."""
        _ = value
        # do not allow user configurate loop params in case of
        # correction loops closed
        if not self._fofb.connected or self._fofb['LoopState-Sts'] or \
                not self._sofb.connected or self._sofb['LoopState-Sts']:
            self._update_log('ERR:Open correction loops before ')
            self._update_log('ERR:configuring AFC Timing RTM loop.')
            return False
        for dev in self._afcti_devs.values():
            if not dev.connected:
                self._update_log(f'ERR:{dev.devname} disconnected.')
                continue
            for propty_sp, desired_val in self._const.AFCTI_CONFIGS:
                dev[propty_sp] = desired_val
            dev['ClkLockedLtcRst-Cmd'] = 1
        return True

    def cmd_config_hltrigs(self, value):
        """Configure HL triggers according to lock configurations."""
        _ = value
        for trigname, configs in self._const.HLTRIG_2_CONFIG:
            trigdev = self._hltrig_devs[trigname]
            if not trigdev.connected:
                self._update_log(f'ERR:{trigname} disconnected.')
                continue
            for prop_sp, desired_val in configs:
                trigdev[prop_sp] = desired_val
        return True

    def cmd_config_llrf(self, value):
        """Configure LLRF interlock according to lock configurations."""
        _ = value
        if not self._llrf.connected:
            self._update_log(f'ERR:LLRF disconnected.')
            return False
        self._llrf['ILK:BEAM:TRIP:S'] = self._llrf_intlk_state
        self._llrf['ILK:MAN:S'] = self._llrf_intlk_state
        return True

    def cmd_config_bpms(self, value):
        """Configure BPMs according to lock configurations."""
        _ = value
        if not self._orbintlk_dev.connected:
            for dev in self._orbintlk_dev.devices:
                self._update_log(f'ERR:{dev.devname} disconnected.')
            return False

        for name, enbl in self._enable_lists.items():
            self.set_enbllist(name, enbl)

        for name, lim in self._limits.items():
            self.set_intlk_lims(name, lim)

        for dev in self._fambpm_dev.devices:
            for prop, desired_val in self._const.SIBPMLOGTRIG_CONFIGS:
                dev[prop] = desired_val
        return True

    def cmd_config_phytrigs(self, value):
        """Configure physical triggers according to lock configurations."""
        _ = value
        for dev in self._phytrig_devs:
            if not dev.connected:
                self._update_log(f'ERR:{dev.devname} disconnected.')
                continue
            for prop, desired_val in self._const.AFCPHYTRIG_CONFIGS:
                # only lock polarity of other AFC physical triggers than SI BPM
                if not self._check_lock_phytrig_prop(dev, prop):
                    continue
                dev[prop] = desired_val
        return True

    # --- status methods ---

    def _config_fout_rxenbl(self):
        fout2rx = dict()
        for chn in self._ti_mon_devs:
            if 'Fout' not in chn:
                continue
            fout = chn.device_name
            outnam = chn.propty_name
            if not outnam:
                continue
            out = int(outnam[-1])
            rx = fout2rx.get(fout, 0)
            rx += 1 << out
            fout2rx[fout] = rx

        for fout, dev in self._fout_devs.items():
            if fout in self._const.FOUTSFIXED_RXENBL:
                continue
            rxenbl = fout2rx.get(fout, 0)
            self._fout2rxenbl[fout] = rxenbl
            if not self._init:
                continue
            dev['RxEnbl-SP'] = rxenbl
            dev._wait('RxEnbl-RB', rxenbl, timeout=1)
            dev['RxLockedLtcRst-Cmd'] = 1
            if rxenbl:
                dev._wait('RxLockedLtc-Mon', rxenbl, timeout=1)

        return True

    def _get_monitored_devices(self):
        enbllist = self._get_gen_bpm_intlk()
        aux = _np.roll(enbllist, 1)
        subsecs = _np.where(_np.sum(aux.reshape(20, -1), axis=1) > 0)[0]
        subsecs += 1

        tidevs = set()
        tidevs.add(self._evg_dev.devname)
        bpmdevs = set()
        for sub in subsecs:
            # timing
            afcti = f'IA-{sub:02}RaBPM:TI-AMCFPGAEVR'
            tidevs.add(afcti)
            foutout = self._const.trigsrc2fout_map[afcti]
            tidevs.add(foutout)
            fout = _PVName(foutout).device_name
            tidevs.add(fout)
            evgout = _LLTimeSearch.get_evg_channel(foutout)
            tidevs.add(evgout)
            if afcti in self._const.REDUNDANCY_TABLE:
                afctir = self._const.REDUNDANCY_TABLE[afcti]
                tidevs.add(afctir)
                foutoutr = self._const.trigsrc2fout_map[afctir]
                tidevs.add(foutoutr)
                foutr = _PVName(foutoutr).device_name
                tidevs.add(foutr)
                evgoutr = _LLTimeSearch.get_evg_channel(foutoutr)
                tidevs.add(evgoutr)
            # bpm
            bpmdevs.update(self._const.crates_map[afcti])

        return sorted(bpmdevs), sorted(tidevs)

    def _check_ti_devices_status(self, devices):
        for devname in devices:
            devname = _PVName(devname)

            dev = self._evg_dev if 'EVG' in devname else \
                self._fout_devs[devname.device_name] if 'Fout' in devname \
                else self._afcti_devs[int(devname.sub[:2])] \
                if 'AMCFPGA' in devname else None
            if dev is None:
                return True

            if not dev.connected:
                self._update_log(f'ERR:{dev.devname} not connected')
                return False
            elif 'Fout' in devname:
                out = int(devname.propty[-1]) if devname.propty else None
                if out is not None and not _get_bit(dev['RxLockedLtc-Mon'], out):
                    self._update_log(f'ERR:{dev.devname} OUT{out} not locked')
                    return False
            elif 'AMCFPGA' in devname and not dev['RTMClkLockedLtc-Mon']:
                self._update_log(f'ERR:{dev.devname} RTM Clk not locked')
                return False
        return True

    def _get_gen_bpm_intlk(self):
        pos, ang = self._enable_lists['pos'], self._enable_lists['ang']
        return _np.logical_or(pos, ang)

    def _check_valid_bpmconfig(self, config):
        aux = _np.roll(config, 1)
        # check if pairs have the same config
        return not _np.any(_np.diff(aux.reshape(-1, 2), axis=1) != 0)

    def _check_lock_phytrig_prop(self, dev, prop):
        devname = _PVName(dev.devname).device_name
        if devname in self._const.bpm_names:
            return True
        return prop == 'DirPol-Sel'

    def _check_configs(self):
        _t0 = _time.time()

        # bpm status
        value = 0
        if self._orbintlk_dev.connected and self._fambpm_dev.connected:
            dev = self._orbintlk_dev
            # PosEnblSynced
            val = _np.array_equal(
                dev.pos_enable, self._enable_lists['pos'])
            value = _updt_bit(value, 1, not val)
            # AngEnblSynced
            val = _np.array_equal(
                dev.ang_enable, self._enable_lists['ang'])
            value = _updt_bit(value, 2, not val)
            # MinSumEnblSynced
            val = _np.array_equal(
                dev.minsum_enable, self._enable_lists['minsum'])
            value = _updt_bit(value, 3, not val)
            # GlobEnblSynced
            genval = self._get_gen_bpm_intlk() if self._state else \
                _np.zeros(self._const.nr_bpms, dtype=bool)
            val = _np.array_equal(dev.gen_enable, genval)
            value = _updt_bit(value, 4, not val)
            # PosLimsSynced
            okp = True
            for prp in ['pos_x_min', 'pos_x_max', 'pos_y_min', 'pos_y_max']:
                okp &= _np.array_equal(
                    getattr(dev, prp+'_thres'), self._limits[prp])
            value = _updt_bit(value, 5, not okp)
            # AngLimsSynced
            oka = True
            for prp in ['ang_x_min', 'ang_x_max', 'ang_y_min', 'ang_y_max']:
                oka &= _np.array_equal(
                    getattr(dev, prp+'_thres'), self._limits[prp])
            value = _updt_bit(value, 6, not oka)
            # MinSumLimsSynced
            oks = _np.array_equal(dev.minsum_thres, self._limits['minsum'])
            value = _updt_bit(value, 7, not oks)
            # AcqConfigured
            bpms = self._fambpm_dev.devices
            okb = all(d.acq_channel == self._acq_chan for d in bpms)
            okb &= all([d.acq_nrsamples_post == self._acq_spost for d in bpms])
            okb &= all([d.acq_nrsamples_pre == self._acq_spre for d in bpms])
            okb &= all(
                d.acq_repeat == self._const.AcqRepeat.Normal for d in bpms)
            okb &= all(
                d.acq_trigger == self._const.AcqTrigTyp.External for d in bpms)
            okb &= all(
                d.acq_status == self._const.AcqStates.External_Trig for d in bpms)
            value = _updt_bit(value, 8, not okb)
            # LogTrigConfigured
            okl = True
            for bpm in self._fambpm_dev.devices:
                for prp, val in self._const.SIBPMLOGTRIG_CONFIGS:
                    prp_rb = _PVName.from_sp2rb(prp)
                    okl &= bpm[prp_rb] == val
            value = _updt_bit(value, 9, not okl)

        self._bpm_status = value
        self.run_callbacks('BPMStatus-Mon', self._bpm_status)

        # Timing Status
        value = 0
        # EVG
        dev = self._evg_dev
        if dev.connected:
            val = dev['IntlkCtrlEnbl-Sts'] != self._state
            value = _updt_bit(value, 1, val)
            okg = True
            for prp, val in self._const.EVG_CONFIGS:
                prp_rb = _PVName.from_sp2rb(prp)
                okg &= dev[prp_rb] == val
            value = _updt_bit(value, 2, not okg)
        else:
            value = 0b111
        # Fouts
        if all(dev.connected for dev in self._fout_devs.values()):
            okg = True
            for devname, rxenbl in self._fout2rxenbl.items():
                dev = self._fout_devs[devname]
                okg &= dev['RxEnbl-RB'] == rxenbl
                okg &= dev['RxLockedLtc-Mon'] == rxenbl
            value = _updt_bit(value, 4, not okg)
        else:
            value += 0b11 << 3
        # AFC timing
        if all(dev.connected for dev in self._afcti_devs.values()):
            okg = True
            for dev in self._afcti_devs.values():
                for prp, val in self._const.AFCTI_CONFIGS:
                    prp_rb = _PVName.from_sp2rb(prp)
                    okg &= dev[prp_rb] == val
            value = _updt_bit(value, 6, not okg)
        else:
            value += 0b11 << 5
        # AFC Physical triggers
        if all(dev.connected for dev in self._phytrig_devs):
            okg = True
            for dev in self._phytrig_devs:
                for prp, val in self._const.AFCPHYTRIG_CONFIGS:
                    if not self._check_lock_phytrig_prop(dev, prp):
                        continue
                    prp_rb = _PVName.from_sp2rb(prp)
                    okg &= dev[prp_rb] == val
            value = _updt_bit(value, 8, not okg)
        else:
            value += 0b11 << 7
        # HL triggers
        bit = 9
        for trigname, configs in self._const.HLTRIG_2_CONFIG:
            dev = self._hltrig_devs[trigname]
            if dev.connected:
                value = _updt_bit(value, bit+1, bool(dev['Status-Mon']))
                oko = True
                for prp, val in configs:
                    prp_rb = _PVName.from_sp2rb(prp)
                    oko &= dev[prp_rb] == val
                value = _updt_bit(value, bit+2, not oko)
            else:
                value += 0b111 << bit
            bit += 3

        self._timing_status = value
        self.run_callbacks('TimingStatus-Mon', self._timing_status)

        # LLRF Status
        value = (1 << 2) - 1
        dev = self._llrf
        if dev.connected:
            value = _updt_bit(value, 0, 0)
            okc = dev['ILK:BEAM:TRIP'] == self._llrf_intlk_state
            okc &= dev['ILK:MAN'] == self._llrf_intlk_state
            value = _updt_bit(value, 1, not okc)
        self.run_callbacks('LLRFStatus-Mon', value)

        # check time elapsed
        ttook = _time.time() - _t0
        tplanned = self.thread_check_configs.interval
        tsleep = tplanned - ttook
        if tsleep <= 0:
            _log.warning(
                'Configuration check took more than planned... '
                '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))

    # --- interlock methods ---

    def _callback_evg_intlk(self, value, **kws):
        _ = kws
        if not self._state:
            return
        if self._thread_cbevgilk and self._thread_cbevgilk.is_alive():
            return
        self._thread_cbevgilk = _CAThread(
            target=self._do_callback_evg_intlk, args=(value, ), daemon=True)
        self._thread_cbevgilk.start()

    def _do_callback_evg_intlk(self, value):
        if value == 0:
            return

        self._update_log('FATAL:Orbit interlock raised by EVG.')

        self._update_log('Waiting a little before rearming...')
        _time.sleep(self._const.DEF_TIME2WAIT_INTLKREARM)

        # reset latch flags for BPM interlock core and EVG
        self.cmd_reset('all')

        # reconfigure BPM configuration
        self.cmd_acq_config()

    def _callback_fout_rxlock(self, pvname, value, **kws):
        if not self._init:
            return
        nam = _PVName(pvname).device_name
        if self._thread_cbfout[nam] and self._thread_cbfout[nam].is_alive():
            return
        self._thread_cbfout[nam] = _CAThread(
            target=self._do_callback_rxlock,
            args=(pvname, value, ), daemon=True)
        self._thread_cbfout[nam].start()

    def _callback_evg_rxlock(self, pvname, value, **kws):
        if not self._init:
            return
        pvname = _PVName(pvname)
        configs = self._const.EVG_CONFIGS
        bits = [int(c[0][-1]) for c in configs if 'RxEnbl' in c[0]]
        if all([_get_bit(value, b) for b in bits]):  # all ok
            return
        if self._thread_cbevgrx and self._thread_cbevgrx.is_alive():
            return
        self._thread_cbevgrx = _CAThread(
            target=self._do_callback_rxlock,
            args=(pvname, value, ), daemon=True)
        self._thread_cbevgrx.start()

    def _do_callback_rxlock(self, pvname, value):
        pvname = _PVName(pvname)
        devname = pvname.device_name
        if pvname.dev == 'EVG':
            is_failure = False
            for bit in range(8):
                if _get_bit(value, bit):
                    continue
                outnam = f'OUT{bit}'
                self._update_log(f'WARN:{outnam} of {devname} not locked')
                devout = devname.substitute(propty_name=outnam)
                # verify if this is an orbit interlock reliability failure
                is_failure |= devout in self._ti_mon_devs
        else:
            outs_in_failure = set()
            for dev in self._ti_mon_devs:
                # verify fouts
                if 'Fout' not in dev:
                    continue
                dev = _PVName(dev)
                # if the fout from callback is a monitored one
                if dev.device_name == devname:
                    # verify if the monitored outs are locked
                    outnam = dev.propty_name
                    if not outnam:
                        continue
                    out = int(outnam[-1])
                    if _get_bit(value, out):
                        continue
                    # if not, it is a reliability failure
                    outs_in_failure.add(out)
                    self._update_log(f'WARN:{outnam} of {devname} not locked')
            # verify redundancy pairs, failure only if both are in failure
            aux_var_loop = outs_in_failure.copy()
            for out in aux_var_loop:
                outnam = f'OUT{out}'
                devout = devname.substitute(propty_name=outnam)
                if devout in self._const.intlkr_fouttable:
                    pair = self._const.intlkr_fouttable[devout]
                    devpair = _PVName(pair).device_name
                    if self._fout_devs[devpair]['RxLockedLtc-Mon']:
                        outs_in_failure.remove(out)
                    else:
                        self._update_log(f'WARN:{outnam} of {pair} not locked')
            is_failure = bool(outs_in_failure)

        if not is_failure:
            return
        self._handle_reliability_failure()

    def _conn_callback_timing(self, pvname, conn, **kws):
        if conn:
            return
        devname = _PVName(pvname).device_name
        # verify if this is an orbit interlock reliability failure
        is_failure = devname in self._ti_mon_devs
        flag = 'FATAL' if is_failure else 'WARN'
        self._update_log(f'{flag}:{devname} disconnected')
        if is_failure:
            self._handle_reliability_failure()

    def _callback_bpm_intlk(self, pvname, value, **kws):
        _ = kws
        if not value:
            return
        # launch thread to log interlock details
        _CAThread(
            target=self._log_bpm_intlk,
            args=(_PVName(pvname).device_name, ),
            daemon=True).start()
        # launch thread to send interlock to RF as a backup
        if self._thread_cbbpm and self._thread_cbbpm.is_alive():
            return
        self._thread_cbbpm = _CAThread(
            target=self._do_callback_bpm_intlk, daemon=True)
        self._thread_cbbpm.start()

    def _log_bpm_intlk(self, bpmname):
        # log which interlock flag was raised
        self._update_log(f'FATAL:{bpmname} raised interlock.')
        props = [
            'IntlkPosLowerLtcX-Mon', 'IntlkPosUpperLtcX-Mon',
            'IntlkPosLowerLtcY-Mon', 'IntlkPosUpperLtcY-Mon',
            'IntlkAngLowerLtcX-Mon', 'IntlkAngUpperLtcX-Mon',
            'IntlkAngLowerLtcY-Mon', 'IntlkAngUpperLtcY-Mon',
        ]
        for prop in props:
            idx = self._const.bpm_names.index(bpmname)
            intlk, pln = prop.split('-')[0].split('Intlk')[1].split('Ltc')
            if self._orbintlk_dev.devices[idx][prop]:
                self._update_log(f'FATAL:{bpmname} > {intlk} {pln}')

    def _do_callback_bpm_intlk(self):
        # send kill beam as fast as possible
        self._handle_reliability_failure(is_failure=False)
        # wait minimum period for RF EVE event count to be updated
        _time.sleep(.1)
        # verify if RF EVE counted the event PsMtm
        new_evtcnt = self._everf_dev[self._llrf_evtcnt_pvname]
        if new_evtcnt == self._everf_evtcnt:
            self._update_log('WARN:RF EVE did not count event PsMtm')
        self._everf_evtcnt = new_evtcnt
        # wait minimum period for BPM to update interlock PVs
        _time.sleep(2)
        # verify if EVG propagated the event Intlk
        evgintlksts = self._evg_dev['IntlkEvtStatus-Mon']
        if not evgintlksts & 0b1:
            self._update_log('ERR:EVG did not propagate event Intlk')
            # reset BPM orbit interlock, once EVG callback was not triggered
            self.cmd_reset('bpm_all')
        if not self._llrf['FASTINLK-MON'] & (1 << 12):
            self._update_log('ERR:LLRF did not received RFKill event')

    def _get_bpm_rates_factor(self):
        if self._monitsum2intlksum_factor:
            return self._monitsum2intlksum_factor
        monit = self._fambpm_dev.devices[0]['INFOMONITRate-RB']
        facq = self._fambpm_dev.devices[0]['INFOFAcqRate-RB']

        if None in [monit, facq]:
            return 0
        frac = monit/facq
        factor = 2**_np.ceil(_np.log2(frac)) / frac
        self._monitsum2intlksum_factor = factor
        return self._monitsum2intlksum_factor

    def _callback_bpm_adclock(self, pvname, value, **kws):
        _ = kws
        if value == 1:
            return
        devname = _PVName(pvname).device_name
        is_failure = devname in self._bpm_mon_devs and \
            devname in self._const.bpm_names
        flag = 'FATAL' if is_failure else 'WARN'
        self._update_log(f'{flag}:{devname} lost PLL lock')
        if is_failure:
            self._handle_reliability_failure()

    def _conn_callback_afcphystrigs(self, pvname, conn, **kws):
        _ = kws
        if conn:
            return
        devname = _PVName(pvname).device_name
        is_failure = devname in self._bpm_mon_devs
        flag = 'ERR' if is_failure else 'WARN'
        self._update_log(f'{flag}:{devname} disconnected')
        if is_failure:
            self._handle_reliability_failure()

    def _callback_hltrig_status(self, pvname, value, **kws):
        _ = kws
        if not value:
            return
        # if status is not ok, it is a reliability failure
        trigname = _PVName(pvname).device_name
        self._update_log(f'FATAL:{trigname} Status not ok')
        self._handle_reliability_failure()

    # --- reliability failure methods ---

    def _check_minsum_requirement(self, monit_sum=None):
        if monit_sum is None:
            monit_sum = self._sofb['SlowSumRaw-Mon']
        facq_sum = monit_sum * self._get_bpm_rates_factor()
        return _np.all(facq_sum > self._limits['minsum'])

    def _handle_reliability_failure(self, is_failure=True):
        if is_failure:
            flag = 'FATAL' if self._state else 'WARN'
            self._update_log(f'{flag}:Orbit interlock reliability failure')
        if not self._state:
            self._update_log('WARN:Orbit interlock is not enabled.')
            return
        # send soft interlock to RF
        self._update_log('FATAL:sending soft interlock to LLRF.')
        self._llrf['IntlkSet-Cmd'] = 1
        _time.sleep(1)
        self._llrf['IntlkSet-Cmd'] = 0
        if self._is_dry_run:
            # wait a little and rearming FDL acquisition
            _time.sleep(self._const.DEF_TIME2WAIT_INTLKREARM)
            self._llrf['Reset-Cmd'] = 1
            _time.sleep(1)
            self._llrf['Reset-Cmd'] = 0

    # --- device lock methods ---

    def _callback_lock(
            self, device, propty_sp, desired_value, pvname, value, **kwargs):
        thread = _CAThread(
            target=self._start_lock_thread,
            args=(device, propty_sp, desired_value, pvname, value),
            daemon=True)
        thread.start()

    def _callback_evg_lock_intlk(self, pvname, value, **kwargs):
        thread = _CAThread(
            target=self._start_lock_thread,
            args=(
                self._evg_dev, 'IntlkCtrlEnbl-Sel', self._state,
                pvname, value),
            daemon=True)
        thread.start()

    def _callback_fout_lock(self, pvname, value, **kwargs):
        devname = _PVName(pvname).device_name
        desired_value = self._fout2rxenbl[devname]
        device = self._fout_devs[devname]
        thread = _CAThread(
            target=self._start_lock_thread,
            args=(device, 'RxEnbl-SP', desired_value, pvname, value),
            daemon=True)
        thread.start()

    def _callback_bpm_lock(self, pvname, value, **kws):
        pvname = _PVName(pvname)
        devname = pvname.device_name
        propty_rb = pvname.propty
        propty_sp = _PVName.from_rb2sp(propty_rb)
        devidx = self._orbintlk_dev.BPM_NAMES.index(devname)
        device = self._orbintlk_dev.devices[devidx]
        if propty_rb.endswith('En-Sts'):
            entyp = 'pos' if 'Pos' in propty_rb else \
                'ang' if 'Ang' in propty_rb else \
                'minsum' if 'MinSum' in propty_rb else \
                'gen'
            if entyp == 'gen':
                desired_value = self._get_gen_bpm_intlk()[devidx] \
                    if self._state else 0
            else:
                desired_value = self._enable_lists[entyp][devidx]
        elif 'Lmt' in propty_rb:
            limcls = 'pos' if 'Pos' in propty_rb else \
                'ang' if 'Ang' in propty_rb else 'minsum'
            limpln = '_x_' if 'X' in propty_rb else \
                '_y_' if 'Y' in propty_rb else ''
            limtyp = '' if 'MinSum' in propty_rb \
                else 'max' if 'Max' in propty_rb else 'min'
            limname = f'{limcls}{limpln}{limtyp}'
            desired_value = self._limits[limname][devidx]

        thread = _CAThread(
            target=self._start_lock_thread,
            args=(device, propty_sp, desired_value, pvname, value),
            daemon=True)
        thread.start()

    def _start_lock_thread(
            self, device, propty_sp, desired_value, pvname, value):
        if self._lock_suspend:
            return

        # do not try to lock devices that are not in list of monitored devices
        devname = _PVName(pvname).device_name
        if devname not in self._ti_mon_devs and \
                devname not in self._bpm_mon_devs and \
                devname not in self._hltrig_devs:
            return

        # if there is already a lock thread, return
        thread = self._lock_threads.get(pvname, None)
        if thread is not None and thread.is_alive():
            return

        # else, create lock thread with 10 attempts to lock PV
        interval = 1 / 10  # little sleep to avoid CPU load
        thread = _Repeat(
            interval, self._do_lock,
            args=(device, propty_sp, desired_value, pvname, value),
            niter=10, is_cathread=True)
        self._lock_threads[pvname] = thread
        thread.start()

    def _do_lock(self, device, propty_sp, desired_value, pvname, value):
        thread = self._lock_threads[pvname]

        # if value is equal desired, stop thread
        if value == desired_value:
            if pvname in self._lock_failures:
                self._lock_failures.remove(pvname)
            thread.stop()
            return

        # else, apply value as desired
        if device == self._llrf:
            propty_rb = propty_sp.replace(':S', '')
        else:
            propty_rb = _PVName.from_sp2rb(propty_sp)
        self._update_log(f'WARN:Locking {pvname}')
        device[propty_sp] = desired_value

        # if readback reached desired value, stop thread
        if device._wait(propty_rb, desired_value, timeout=0.11):
            if pvname in self._lock_failures:
                self._lock_failures.remove(pvname)
            thread.stop()
            return

        # if this was the last iteration, raise a reliability failure
        if thread.cur_iter == thread.niters-1:
            self._lock_failures.add(pvname)
            self._update_log(f'FATAL:Fail to lock {pvname}')
            self._handle_reliability_failure()

    # --- auxiliary log methods ---

    def _update_log(self, msg):
        if 'ERR' in msg:
            _log.error(msg[4:])
        elif 'FATAL' in msg:
            _log.error(msg[6:])
        elif 'WARN' in msg:
            _log.warning(msg[5:])
        else:
            _log.info(msg)
        self.run_callbacks('Log-Mon', msg)

    # --- file handlers ---

    def _load_file(self, intlk, dtype='en'):
        filename, desc = self._get_file_info(intlk, dtype)
        if not _os.path.isfile(filename):
            return
        value = _np.loadtxt(filename)
        okl = True
        if dtype.startswith('en'):
            okl = self.set_enbllist(intlk, value)
        elif dtype.startswith('lim'):
            okl = self.set_intlk_lims(intlk, value)
        if okl:
            msg = f'Loaded {intlk} {desc} from auto save!'
        else:
            msg = f'ERR:Problem loading {intlk} {desc} from file.'
        self._update_log(msg)
        return okl

    def _save_file(self, intlk, value, dtype):
        filename, desc = self._get_file_info(intlk, dtype)
        try:
            path = _os.path.split(filename)[0]
            _os.makedirs(path, exist_ok=True)
            _np.savetxt(filename, value)
        except FileNotFoundError:
            self._update_log(
                f'WARN:Could not save {intlk} {desc} to file.')

    def _get_file_info(self, intlk, dtype):
        if dtype.startswith(('en', 'lim')):
            desc = 'enable list' if dtype.startswith('en') else 'limits'
            suff = '_enbl' if dtype.startswith('en') else '_lim'
            fname = intlk + suff
        else:
            raise ValueError(f'file info not defined for {intlk} and {dtype}')
        filename = getattr(self._const, fname + '_fname')
        return filename, desc



#!/usr/bin/env python-sirius
"""Test webserver implementation module."""
from unittest import TestCase, mock
from urllib.request import URLError

from siriuspy.clientweb import implementation
import siriuspy.util as util

_TIMEOUT = implementation._TIMEOUT
# Dependencies
# _envars.SRVURL_CSCONSTS
# _urllib_request.urlopen
# _urllib_request.urlopen.read


class TestClientWebReadUrl(TestCase):
    """Test read url method."""

    def setUp(self):
        """Common setup for all tests."""
        # Create mocks
        url_patcher = mock.patch.object(
            implementation, '_urllib_request', autospec=True)
        env_patcher = mock.patch.object(
            implementation, '_envars', autospec=True)
        self.addCleanup(url_patcher.stop)
        self.addCleanup(env_patcher.stop)
        self.url_mock = url_patcher.start()
        self.env_mock = env_patcher.start()
        # Configuration parameters
        self.fake_url = "http://FakeBaseURL/"
        # Mocked methods
        self.url_mock.urlopen.return_value.read.return_value = b'FakeResponse'
        # Mock a property from envars
        type(self.env_mock).SRVURL_CSCONSTS = \
            mock.PropertyMock(return_value=self.fake_url)

    def _test_read_url_request(self):
        """Test read_rul makes a request using urlopen."""
        implementation.read_url("FakeURL")
        self.url_mock.urlopen.assert_called_once_with(
            self.fake_url + "FakeURL", timeout=1.0)

    def _test_read_url_response(self):
        """Test read_url."""
        self.assertEqual(implementation.read_url("FakeURL"), "FakeResponse")

    def _test_read_url_exception(self):
        """Test read_url raises exception."""
        self.url_mock.urlopen.side_effect = URLError("FakeError")
        with self.assertRaises(Exception):
            implementation.read_url("FakeURL")


@mock.patch.object(implementation, 'read_url', autospec=True,
                   return_value="FakeResponse")
class TestClientWeb(TestCase):
    """Test ClientWeb."""

    public_interface = {
        'read_url',
        'server_online',
        'magnets_model_data',
        'magnets_excitation_data_read',
        'magnets_excitation_ps_read',
        'ps_pstypes_names_read',
        'ps_pstype_data_read',
        'ps_pstype_setpoint_limits',
        'pu_pstype_setpoint_limits',
        'ps_siggen_configuration_read',
        'ps_psmodels_read',
        'pu_psmodels_read',
        'beaglebone_freq_mapping',
        'beaglebone_ip_list',
        'bbb_udc_mapping',
        'udc_ps_mapping',
        'crates_mapping',
        'bpms_data',
        'timing_devices_mapping',
        'high_level_triggers',
        'high_level_events',
        'bsmp_dclink_mapping',
        'mac_schedule_read',
        'doc_services_read',
    }

    def test_public_interface(self, mock_read):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            implementation, TestClientWeb.public_interface)
        self.assertTrue(valid)

    def test_server_online(self, mock_read):
        """Test server_online return True when no exception is issued."""
        self.assertTrue(implementation.server_online())

    def test_server_online_exception(self, mock_read):
        """Test server_online return True when no exception is issued."""
        mock_read.side_effect = Exception()
        self.assertFalse(implementation.server_online())

    # def _test_bpms_data(self, mock_read):
    #     """Test bpms_data."""
    #     url = implementation._magnet_folder + 'magnets-model-data.txt'
    #     # Call with different parameters
    #     resp = implementation.bpms_data()
    #     self.assertEqual(resp, "FakeResponse")
    #     resp = implementation.bpms_data(timeout=2.0)
    #     self.assertEqual(resp, "FakeResponse")
    #     # Assert read_url was called correctly
    #     mock_read.assert_has_calls([
    #         mock.call(url, timeout=1.0),
    #         mock.call(url, timeout=2.0)])

    def test_magnets_excitation_data_read(self, mock_read):
        """Test magnets_excitation_data_read."""
        filename = "fakefile"
        folder = implementation._EXCDAT_FOLDER
        # Call two times
        resp = implementation.magnets_excitation_data_read(filename)
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.magnets_excitation_data_read(
            filename, timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read was called correctly
        mock_read.assert_has_calls([
            mock.call(folder + filename, timeout=_TIMEOUT),
            mock.call(folder + filename, timeout=2*_TIMEOUT)])

    def test_magnets_excitation_ps_read(self, mock_read):
        """Test magnets_excitation_ps_read."""
        url = implementation._MAGNET_FOLDER + 'magnet-excitation-ps.txt'
        # Call with different parameters
        resp = implementation.magnets_excitation_ps_read()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.magnets_excitation_ps_read(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_ps_pstypes_names_read(self, mock_read):
        """Test ps_pstypes_names_read."""
        url = implementation._PS_FOLDER + 'pstypes-names.txt'
        # Call with different parameters
        resp = implementation.ps_pstypes_names_read()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.ps_pstypes_names_read(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_ps_pstype_data_read(self, mock_read):
        """Test ps_pstype_data_read."""
        filename = "fakefilename"
        url = implementation._PSTYPES_DATA_FOLDER + filename
        # Call with different parameters
        resp = implementation.ps_pstype_data_read(filename)
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.ps_pstype_data_read(filename, timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_ps_siggen_configuration_read(self, mock_read):
        """Test ps_siggen_configuration_read."""
        url = implementation._PS_FOLDER + 'siggen-configuration.txt'
        # Call with different parameters
        resp = implementation.ps_siggen_configuration_read()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.ps_siggen_configuration_read(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_ps_pstype_setpoint_limits(self, mock_read):
        """Test ps_pstype_setpoint_limits."""
        url = implementation._PS_FOLDER + 'pstypes-setpoint-limits.txt'
        # Call with different parameters
        resp = implementation.ps_pstype_setpoint_limits()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.ps_pstype_setpoint_limits(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_pu_pstype_setpoint_limits(self, mock_read):
        """Test pu_pstype_setpoint_limits."""
        url = implementation._PS_FOLDER + 'putypes-setpoint-limits.txt'
        # Call with different parameters
        resp = implementation.pu_pstype_setpoint_limits()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.pu_pstype_setpoint_limits(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_beaglebone_freq_mapping(self, mock_read):
        """Test beaglebone_freqs_mapping."""
        # TODO: implement!
        pass

    def test_beaglebone_ip_list(self, mock_read):
        """Test beaglebone_ip_list."""
        # TODO: implement!
        pass

    def test_crates_mapping(self, mock_read):
        """Test crate_to_bpm_mapping."""
        url = (
            implementation._DIAG_FOLDER + 'microTCA-vs-BPMs-mapping/')
        # Call with different parameters
        resp = implementation.crates_mapping()
        self.assertEqual(resp, "")
        resp = implementation.crates_mapping(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_bpms_data(self, mock_read):
        """Test bpms_data."""
        url = implementation._DIAG_FOLDER + 'bpms-data.txt'
        # Call with different parameters
        resp = implementation.bpms_data()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.bpms_data(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_timing_devices_mapping(self, mock_read):
        """Test timing_devices_mapping."""
        url = implementation._TIMESYS_FOLDER + 'timing-devices-connection.txt'
        # Call with different parameters
        resp = implementation.timing_devices_mapping()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.timing_devices_mapping(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_high_level_triggers(self, mock_read):
        """Test high_level_triggers."""
        url = implementation._TIMESYS_FOLDER + 'high-level-triggers.py'
        # Call with different parameters
        resp = implementation.high_level_triggers()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.high_level_triggers(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])

    def test_high_level_events(self, mock_read):
        """Test high_level_events."""
        url = implementation._TIMESYS_FOLDER + 'high-level-events.py'
        # Call with different parameters
        resp = implementation.high_level_events()
        self.assertEqual(resp, "FakeResponse")
        resp = implementation.high_level_events(timeout=2*_TIMEOUT)
        self.assertEqual(resp, "FakeResponse")
        # Assert read_url was called correctly
        mock_read.assert_has_calls([
            mock.call(url, timeout=_TIMEOUT),
            mock.call(url, timeout=2*_TIMEOUT)])



"""Ramp module with conversion classes."""

from .. import util as _util
from ..namesys import SiriusPVName as _SiriusPVName
from ..search import PSSearch as _PSSearch, MASearch as _MASearch
from ..magnet import util as _mutil
from ..magnet.factory import NormalizerFactory as _NormalizerFactory


from .exceptions import RampInvalidDipoleWfmParms as \
    _RampInvalidDipoleWfmParms


_MAGNETS_DICT = dict()


def get_magnet(maname):
    """."""
    if maname not in _MAGNETS_DICT:
        _MAGNETS_DICT[maname] = Magnet(maname)
    return _MAGNETS_DICT[maname]


class Magnet:
    """Base class for handling magnets."""

    def __init__(self, maname):
        """Init method."""
        self._maname = _SiriusPVName(maname)
        psnames = _MASearch.conv_maname_2_psnames(maname)
        pstype = _PSSearch.conv_psname_2_pstype(psnames[0])
        self._magfunc = _PSSearch.conv_pstype_2_magfunc(pstype)
        self._splims = _PSSearch.conv_psname_2_splims(psnames[0])
        self._dipole_name = _mutil.get_section_dipole_name(self._maname)
        self._family_name = _mutil.get_magnet_family_name(self._maname)
        self._strength_obj = _NormalizerFactory.create(self._maname)
        self._strength_label = _util.get_strength_label(self._magfunc)
        self._strength_units = _util.get_strength_units(self._magfunc)

        if maname not in _MAGNETS_DICT:
            _MAGNETS_DICT[maname] = self

    @property
    def maname(self):
        """Return magnet device name."""
        return self._maname

    @property
    def section(self):
        """Section the magnet belongs to."""
        return self._maname.sec

    @property
    def dipole_name(self):
        """Return section dipole device name."""
        return self._dipole_name

    @property
    def family_name(self):
        """Return section dipole device name."""
        return self._family_name

    @property
    def magfunc(self):
        """Return magnet function."""
        return self._magfunc

    @property
    def strength_label(self):
        """Return strength label of magnet."""
        return self._strength_label

    @property
    def strength_units(self):
        """Strength units of magnet."""
        return self._strength_units

    @property
    def current_min(self):
        """Mininum current for magnet power supply."""
        return self._splims['DRVL']

    @property
    def current_max(self):
        """Maximum current for magnet power supply."""
        return self._splims['DRVH']

    @property
    def splims(self):
        """Magnet SP limits."""
        return self._splims

    def conv_current_2_strength(self, currents, **kwargs):
        """Return strength value from current(s)."""
        strength = self._strength_obj.conv_current_2_strength(currents,
                                                              **kwargs)
        return strength

    def conv_strength_2_current(self, strengths, **kwargs):
        """Return current value from strengths(s)."""
        try:
            current = self._strength_obj.conv_strength_2_current(strengths,
                                                                 **kwargs)
        except ValueError:
            raise _RampInvalidDipoleWfmParms(
                'Could not convert current to energy')
        return current

    def __str__(self):
        """__str__ method."""
        rst = ''
        rst += '\nmaname         : {}'.format(self.maname)
        rst += '\nsection        : {}'.format(self.section)
        rst += '\ndipole_name    : {}'.format(self.dipole_name)
        rst += '\nfamily_name    : {}'.format(self.family_name)
        rst += '\nmagfunc        : {}'.format(self.magfunc)
        rst += '\nstrength_label : {}'.format(self.strength_label)
        rst += '\nstrength_units : {}'.format(self.strength_units)
        rst += '\ncurrent_min    : {}'.format(self.current_min)
        rst += '\ncurrent_max    : {}'.format(self.current_max)
        return rst



"""Define Insertion Devices."""

import time as _time
import inspect as _inspect

from ..search import IDSearch as _IDSearch

from .device import Device as _Device


class _PARAM_PVS:
    """."""

    # --- GENERAL ---
    PERIOD_LEN_CTE = None
    START_PARKING_CMD = None
    IS_MOVING = 'Moving-Mon'
    BLCTRL_ENBL_SEL = 'BeamLineCtrlEnbl-Sel'
    BLCTRL_ENBL_STS = 'BeamLineCtrlEnbl-Sts'
    MOVE_ABORT = None

    # --- PPARAM ---
    PPARAM_SP = None
    PPARAM_RB = None
    PPARAM_MON = None
    PPARAM_PARKED_CTE = None
    PPARAM_MAXVELO_SP = None
    PPARAM_MAXVELO_RB = None
    PPARAM_VELO_SP = None
    PPARAM_VELO_RB = None
    PPARAM_VELO_MON = None
    PPARAM_TOL_SP = None
    PPARAM_TOL_RB = None
    PPARAM_CHANGE_CMD = None

    # --- KPARAM ---
    KPARAM_SP = None
    KPARAM_RB = None
    KPARAM_MON = None
    KPARAM_PARKED_CTE = None
    KPARAM_MAXVELO_SP = None
    KPARAM_MAXVELO_RB = None
    KPARAM_VELO_SP = None
    KPARAM_VELO_RB = None
    KPARAM_VELO_MON = None
    KPARAM_TOL_SP = None
    KPARAM_TOL_RB = None
    KPARAM_CHANGE_CMD = None

    # --- POL ---
    POL_SEL = None
    POL_STS = None
    POL_MON = None
    POL_CHANGE_CMD = None

    def __str__(self):
        """Print parameters."""
        str_ = ''
        strf = '{}: {}'
        for key, value in _inspect.getmembers(self):
            if not key.startswith('_') and value is not None:
                lstr = strf.format(key, value)
                str_ += lstr if str_ == '' else '\n' + lstr
        return str_


class _ID(_Device):
    """Generic Insertion Device."""

    _SHORT_SHUT_EYE = 0.1  # [s]
    _DEF_TIMEOUT = 8  # [s]

    PARAM_PVS = _PARAM_PVS()

    PROPERTIES_DEFAULT = \
        tuple(set(value for key, value in _inspect.getmembers(PARAM_PVS) \
            if not key.startswith('_') and value is not None))

    def __init__(self, devname, props2init='all', auto_monitor_mon=True):
        """."""
        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)
        self._pols_sel_str = \
            _IDSearch.conv_idname_2_polarizations(self.devname)
        self._pols_sts_str = \
            _IDSearch.conv_idname_2_polarizations_sts(self.devname)

    # --- general ---

    @property
    def parameters(self):
        """Return ID parameters."""
        return _IDSearch.conv_idname_2_parameters(self.devname)

    @property
    def period_length(self):
        """Return ID period length [mm]."""
        if self.PARAM_PVS.PERIOD_LEN_CTE in self.properties_all:
            return self[self.PARAM_PVS.PERIOD_LEN_CTE]
        else:
            return self.parameters.PERIOD_LENGTH

    # --- polarization ---

    @property
    def polarization(self):
        """Return ID polarization."""
        if self.PARAM_PVS.POL_STS in self.properties_all:
            return self[self.PARAM_PVS.POL_STS]
        else:
            if _IDSearch.POL_UNDEF_STR in self._pols_sts_str:
                return self._pols_sts_str.index(_IDSearch.POL_UNDEF_STR)
            else:
                return None

    @property
    def polarization_str(self):
        """Return ID polarization string."""
        pol_idx = self.polarization
        if pol_idx is None:
            return None
        else:
            return self._pols_sts_str[pol_idx]

    @polarization.setter
    def polarization(self, value):
        """Set ID polarization."""
        if self.PARAM_PVS.POL_SEL in self.properties_all:
            if isinstance(value, str):
                value = self._pols_sel_str.index(value)
            self[self.PARAM_PVS.POL_SEL] = value

    @property
    def polarization_mon(self):
        """Return ID polarization monitor."""
        if self.PARAM_PVS.POL_MON in self.properties_all:
            return self[self.PARAM_PVS.POL_MON]
        else:
            return None

    @property
    def polarization_mon_str(self):
        """Return ID polarization string."""
        pol_idx = self.polarization_mon
        if pol_idx is None:
            return None
        else:
            return self._pols_sts_str[pol_idx]

    # --- pparameter ---

    @property
    def pparameter_tol(self):
        """PParameter position tolerance [mm]."""
        if self.PARAM_PVS.PPARAM_TOL_RB is None:
            return self.parameters.PPARAM_TOL
        else:
            return self[self.PARAM_PVS.PPARAM_TOL_RB]

    @property
    def pparameter_parked(self):
        """Return ID parked pparameter value [mm]."""
        if self.PARAM_PVS.PPARAM_PARKED_CTE in self.properties_all:
            return self[self.PARAM_PVS.PPARAM_PARKED_CTE]
        else:
            return self.parameters.PPARAM_PARKED

    @property
    def pparameter_speed_max(self):
        """Return max pparameter speed readback [mm/s]."""
        if self.PARAM_PVS.PPARAM_MAXVELO_RB is None:
            return None
        else:
            return self[self.PARAM_PVS.PPARAM_MAXVELO_RB]

    @property
    def pparameter_speed_max_lims(self):
        """Return max pparameter speed limits."""
        if self.PARAM_PVS.PPARAM_MAXVELO_RB is None:
            return None
        else:
            ctrl = self.pv_ctrlvars(self.PARAM_PVS.PPARAM_MAXVELO_SP)
            lims = [ctrl['lower_ctrl_limit'], ctrl['upper_ctrl_limit']]
            return lims

    @property
    def pparameter_speed(self):
        """Return pparameter speed readback [mm/s]."""
        if self.PARAM_PVS.PPARAM_VELO_RB is None:
            return None
        else:
            return self[self.PARAM_PVS.PPARAM_VELO_SP]

    @property
    def pparameter_lims(self):
        """Return ID pparameter lower control limit [mm]."""
        if self.PARAM_PVS.PPARAM_VELO_SP is None:
            return None
        else:
            ctrl = self.pv_ctrlvars(self.PARAM_PVS.PPARAM_SP)
            return [ctrl['lower_ctrl_limit'], ctrl['upper_ctrl_limit']]

    @property
    def pparameter(self):
        """Return ID pparameter readback [mm]."""
        if self.PARAM_PVS.PPARAM_RB is None:
            return None
        else:
            return self[self.PARAM_PVS.PPARAM_RB]

    @property
    def pparameter_mon(self):
        """Return ID pparameter monitor [mm]."""
        if self.PARAM_PVS.PPARAM_MON is None:
            return None
        else:
            return self[self.PARAM_PVS.PPARAM_MON]

    def pparameter_set(self, pparam, timeout=None):
        """Set ID target pparameter for movement [mm]."""
        if self.PARAM_PVS.PPARAM_SP is None:
            return True
        else:
            return self._write_sp(self.PARAM_PVS.PPARAM_SP, pparam, timeout)

    def pparameter_speed_set(self, pparam_speed, timeout=None):
        """Command to set ID cruise pparameterspeed for movement [mm/s]."""
        if self.PARAM_PVS.PPARAM_VELO_SP is None:
            return True
        else:
            return self._write_sp(
                self.PARAM_PVS.PPARAM_VELO_SP, pparam_speed, timeout)

    def pparameter_speed_max_set(self, pparam_speed_max, timeout=None):
        """Command to set ID max cruise pparam speed for movement [mm/s]."""
        if self.PARAM_PVS.PPARAM_MAXVELO_SP is None:
            return True
        else:
            return self._write_sp(
                self.PARAM_PVS.PPARAM_MAXVELO_SP, pparam_speed_max, timeout)

    # --- kparameter ---

    @property
    def kparameter_tol(self):
        """KParameter position tolerance [mm]."""
        if self.PARAM_PVS.KPARAM_TOL_RB is None:
            return self.parameters.KPARAM_TOL
        else:
            return self[self.PARAM_PVS.KPARAM_TOL_RB]

    @property
    def kparameter_parked(self):
        """Return ID parked kparameter value [mm]."""
        if self.PARAM_PVS.KPARAM_PARKED_CTE in self.properties_all:
            return self[self.PARAM_PVS.KPARAM_PARKED_CTE]
        else:
            return self.parameters.KPARAM_PARKED

    @property
    def kparameter_speed_max(self):
        """Return max kparameter speed readback [mm/s]."""
        return self[self.PARAM_PVS.KPARAM_MAXVELO_RB]

    @property
    def kparameter_speed_max_lims(self):
        """Return max kparameter speed limits."""
        ctrl = self.pv_ctrlvars(self.PARAM_PVS.KPARAM_MAXVELO_SP)
        lims = [ctrl['lower_ctrl_limit'], ctrl['upper_ctrl_limit']]
        return lims

    @property
    def kparameter_speed(self):
        """Return kparameter speed readback [mm/s]."""
        return self[self.PARAM_PVS.KPARAM_VELO_SP]

    @property
    def kparameter_speed_mon(self):
        """Return kparameter speed monitor [mm/s]."""
        if self.PARAM_PVS.KPARAM_VELO_MON in self.properties_all:
            return self[self.PARAM_PVS.KPARAM_VELO_MON]
        else:
            raise TypeError('ID does not have speed_mon PV!')

    @property
    def kparameter_lims(self):
        """Return ID kparameter control limits [mm]."""
        ctrl = self.pv_ctrlvars(self.PARAM_PVS.KPARAM_SP)
        return [ctrl['lower_ctrl_limit'], ctrl['upper_ctrl_limit']]

    @property
    def kparameter(self):
        """Return ID kparameter readback [mm]."""
        return self[self.PARAM_PVS.KPARAM_RB]

    @property
    def kparameter_mon(self):
        """Return ID kparameter monitor [mm]."""
        return self[self.PARAM_PVS.KPARAM_MON]

    def kparameter_set(self, kparam, timeout=None):
        """Set ID target kparameter for movement [mm]."""
        return self._write_sp(self.PARAM_PVS.KPARAM_SP, kparam, timeout)

    def kparameter_speed_set(self, kparam_speed, timeout=None):
        """Command to set ID cruise kparam speed for movement [mm/s]."""
        return self._write_sp(
            self.PARAM_PVS.KPARAM_VELO_SP, kparam_speed, timeout)

    def kparameter_speed_max_set(self, kparam_speed_max, timeout=None):
        """Command to set ID max cruise kparam speed for movement [mm/s]."""
        return self._write_sp(
            self.PARAM_PVS.KPARAM_MAXVELO_SP, kparam_speed_max, timeout)

    # --- checks ---

    @property
    def is_moving(self):
        """Return True if phase is changing."""
        return round(self['Moving-Mon']) == 1

    @property
    def is_beamline_ctrl_enabled(self):
        """Return beamline control enabled state (True|False)."""
        return self['BeamLineCtrlEnbl-Sts'] != 0

    # --- cmd_beamline and cmd_drive

    def cmd_beamline_ctrl_enable(self, timeout=None):
        """Command enable bealine ID control."""
        return self._write_sp('BeamLineCtrlEnbl-Sel', 1, timeout)

    def cmd_beamline_ctrl_disable(self, timeout=None):
        """Command disable bealine ID control."""
        return self._write_sp('BeamLineCtrlEnbl-Sel', 0, timeout)

    # --- wait

    def wait_while_busy(self, timeout=None):
        """Command wait within timeout while ID control is busy."""
        return True

    def wait_move_start(self, timeout=None):
        """Wait for movement to start."""
        return self._wait('Moving-Mon', 1, timeout)

    def wait_move_finish(self, timeout=None):
        """Wait for movement to finish."""
        return self._wait('Moving-Mon', 0, timeout)

    def wait_move_config(self, pparam, kparam, timeout):
        """."""
        tol_kparam, tol_pparam = self.kparameter_tol, self.pparameter_tol
        # wait for movement within reasonable time
        time_init = _time.time()
        while True:
            k_pos_ok = True if kparam is None else \
                abs(abs(self.kparameter_mon) - abs(kparam)) <= tol_kparam
            p_pos_ok = True if pparam is None else \
                abs(self.pparameter_mon - pparam) <= tol_pparam
            if p_pos_ok and k_pos_ok and not self.is_moving:
                return True
            if _time.time() - time_init > timeout:
                print(f'tol_total: {timeout:.3f} s')
                print(f'wait_time: {_time.time() - time_init:.3f} s')
                print()
                return False
            _time.sleep(self._SHORT_SHUT_EYE)

    # --- cmd_move ---

    def cmd_move_disable(self):
        """Command to disable and break ID movements."""
        return True

    def cmd_move_enable(self):
        """Command to enable ID movements."""
        return True

    def cmd_move_abort(self):
        """."""
        if self.PARAM_PVS.MOVE_ABORT is not None:
            self[self.PARAM_PVS.MOVE_ABORT] = 1

    def cmd_move_stop(self, timeout=None):
        """Command to interrupt and then enable phase movements."""
        timeout = timeout or self._DEF_TIMEOUT

        # wait for not busy state
        if not self.wait_while_busy(timeout=timeout):
            return False

        # send abort command
        self.cmd_move_abort()

        # send disable command
        self.cmd_move_disable()

        # check for successful stop
        if not self.wait_while_busy(timeout=timeout):
            return False
        if not self.wait_move_finish(timeout=timeout):
            return False

        # enable movement again
        if not self.cmd_move_enable():
            return False

        return True

    def cmd_move_start(self, timeout=None):
        """Command to start movement."""
        success = True
        success &= self.cmd_move_pparameter_start(timeout=timeout)
        success &= self.cmd_move_kparameter_start(timeout=timeout)
        return success

    def cmd_move_pparameter_start(self, timeout=None):
        """Command to start Pparameter movement."""
        return self._move_start(
            self.PARAM_PVS.PPARAM_CHANGE_CMD, timeout=timeout)

    def cmd_move_kparameter_start(self, timeout=None):
        """Command to start Kparameter movement."""
        return self._move_start(
            self.PARAM_PVS.KPARAM_CHANGE_CMD, timeout=timeout)

    def cmd_change_polarization_start(self, timeout=None):
        """Change polarization."""
        return self._move_start(
            self.PARAM_PVS.POL_CHANGE_CMD, timeout=timeout)

    def cmd_move_park(self, timeout=None):
        """Move ID to parked config."""
        pparam, kparam = self.pparameter_parked, self.kparameter_parked
        timeout = self._calc_eta(pparam, kparam) if timeout is None \
            else timeout
        if self.PARAM_PVS.START_PARKING_CMD is not None:
            self[self.PARAM_PVS.START_PARKING_CMD] = 1
            return self.wait_move_config(pparam, kparam, timeout)
        else:
            return self.cmd_move(pparam, kparam, timeout)

    def cmd_move_pparameter(self, pparam=None, timeout=None):
        """Command to set and start pparam movement."""
        pparam = self.pparameter if pparam is None else pparam
        return self.cmd_move(pparam, None, timeout)

    def cmd_move_kparameter(self, kparam=None, timeout=None):
        """Command to set and start kparam movement."""
        kparam = self.kparameter if kparam is None else kparam
        return self.cmd_move(None, kparam, timeout)

    def cmd_move(self, pparam=None, kparam=None, timeout=None):
        """Command to set and start pparam and kparam movements."""
        if self.PARAM_PVS.PPARAM_SP is None:
            pparam = None

        # set target pparam and kparam
        t0_ = _time.time()
        if pparam is not None and \
                not self.pparameter_set(pparam, timeout=timeout):
            return False
        if kparam is not None and \
                not self.kparameter_set(kparam, timeout=timeout):
            return False
        t1_ = _time.time()
        if timeout is not None:
            timeout = max(timeout - (t1_ - t0_), 0)

        # command move start
        t0_ = _time.time()
        if pparam is not None and \
                not self.cmd_move_pparameter_start(timeout=timeout):
            return False
        if kparam is not None and \
                not self.cmd_move_kparameter_start(timeout=timeout):
            return False
        t1_ = _time.time()
        if timeout is not None:
            timeout = max(timeout - (t1_ - t0_), 0)

        # calc ETA
        eta = self._calc_eta(pparam, kparam)
        timeout = eta if timeout is None else timeout

        # wait for movement within reasonable time
        return self.wait_move_config(pparam, kparam, timeout)

    def cmd_change_polarization(self, polarization, timeout=None):
        """."""
        if self.PARAM_PVS.POL_SEL not in self.properties_all:
            return True
        if self.PARAM_PVS.POL_MON not in self.properties_all:
            return True

        t0_ = _time.time()

        # set desired polarization
        if not self._write_sp(
                self.PARAM_PVS.POL_SEL, polarization, timeout=timeout):
            return False
        t1_ = _time.time()
        timeout = max(0, timeout - (t1_ - t0_))

        # send change polarization command
        if not self.cmd_change_polarization_start(timeout=timeout):
            return False
        t2_ = _time.time()
        timeout = max(0, timeout - (t2_ - t0_))

        # wait for polarization value within timeout
        return self._wait(
            self.PARAM_PVS.POL_MON, polarization, timeout=timeout, comp='eq')

    # --- private methods ---

    def _move_start(self, cmd_propty, timeout=None, cmd_value=1):
        timeout = timeout or self._DEF_TIMEOUT

        # wait for not busy state
        if not self.wait_while_busy(timeout=timeout):
            return False

        if cmd_propty in self.properties_all:
            # send move command
            self[cmd_propty] = cmd_value

        return True

    def _write_sp(self, propties_sp, values, timeout=None, pvs_sp_rb=None):
        timeout = timeout or self._DEF_TIMEOUT
        if isinstance(propties_sp, str):
            propties_sp = (propties_sp, )
            values = (values, )
        success = True
        for propty_sp, value in zip(propties_sp, values):
            if pvs_sp_rb is not None and propty_sp in pvs_sp_rb:
                # pv is unique for SP and RB variables.
                propty_rb = propty_sp
            else:
                propty_rb = \
                    propty_sp.replace('-SP', '-RB').replace('-Sel', '-Sts')
            self[propty_sp] = value
            if isinstance(value, float):
                success &= super()._wait_float(
                    propty_rb, value, timeout=timeout)
            else:
                success &= super()._wait(
                    propty_rb, value, timeout=timeout, comp='eq')
        return success

    def _wait_propty(self, propty, value, timeout=None):
        """."""
        t0_ = _time.time()
        timeout = timeout if timeout is not None else self._DEF_TIMEOUT
        while self[propty] != value:
            _time.sleep(self._SHORT_SHUT_EYE)
            if _time.time() - t0_ > timeout:
                return False
        return True

    def _calc_eta(self, pparam, kparam):
        """."""
        # calc ETA
        dtime_kparam = 0 if kparam is None else \
            abs(kparam - self.kparameter_mon) / self.kparameter_speed
        dtime_pparam = 0 if pparam is None else \
            abs(pparam - self.pparameter_mon) / self.pparameter_speed
        dtime_max = self._calc_eta_select_time(dtime_kparam, dtime_pparam)
        # additional percentual in ETA
        eta = 1.5 * dtime_max + 5
        return eta

    @staticmethod
    def _calc_eta_select_time(dtime_kparam, dtime_pparam):
        return dtime_kparam + dtime_pparam


class APU(_ID):
    """APU Insertion Device."""

    class DEVICES:
        """Device names."""

        APU22_06SB = 'SI-06SB:ID-APU22'
        APU22_07SP = 'SI-07SP:ID-APU22'
        APU22_08SB = 'SI-08SB:ID-APU22'
        APU22_09SA = 'SI-09SA:ID-APU22'
        APU58_11SP = 'SI-11SP:ID-APU58'
        ALL = (
            APU22_06SB, APU22_07SP, APU22_08SB, APU22_09SA, APU58_11SP, )

    _CMD_MOVE_STOP, _CMD_MOVE_START = 1, 3
    _CMD_MOVE = 3

    # --- PARAM_PVS ---
    PARAM_PVS = _PARAM_PVS()
    PARAM_PVS.KPARAM_SP = 'Phase-SP'
    PARAM_PVS.KPARAM_RB = 'Phase-SP'  # There is no Phase-RB!
    PARAM_PVS.KPARAM_MON = 'Phase-Mon'
    PARAM_PVS.KPARAM_MAXVELO_SP = 'MaxPhaseSpeed-SP'
    PARAM_PVS.KPARAM_MAXVELO_RB = 'MaxPhaseSpeed-RB'
    PARAM_PVS.KPARAM_VELO_SP = 'PhaseSpeed-SP'
    PARAM_PVS.KPARAM_VELO_RB = 'PhaseSpeed-SP'
    PARAM_PVS.KPARAM_VELO_MON = 'PhaseSpeed-Mon'
    PARAM_PVS.KPARAM_CHANGE_CMD = 'DevCtrl-Cmd'

    PROPERTIES_DEFAULT = \
        tuple(set(value for key, value in _inspect.getmembers(PARAM_PVS) \
            if not key.startswith('_') and value is not None))

    def __init__(self, devname, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname not in APU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)

    # --- phase speeds ----

    @property
    def phase_speed_max(self):
        """Return max phase speed readback [mm/s]."""
        return self.kparameter_speed_max

    @property
    def phase_speed_max_lims(self):
        """Return max phase speed limits."""
        return self.kparameter_speed_max_lims

    @property
    def phase_speed(self):
        """Return phase speed [mm/s]."""
        return self.kparameter_speed

    @property
    def phase_speed_mon(self):
        """Return phase speed monitor [mm/s]."""
        return self.kparameter_speed_mon

    # --- phase ---

    @property
    def phase_parked(self):
        """Return ID parked phase value [mm]."""
        return self.kparameter_parked

    @property
    def phase_lims(self):
        """Return ID phase lower control limit [mm]."""
        return self.kparameter_lims

    @property
    def phase(self):
        """Return APU phase [mm]."""
        return self.kparameter

    @property
    def phase_mon(self):
        """Return APU phase [mm]."""
        return self.kparameter_mon

    # --- set methods ---

    def phase_set(self, phase, timeout=None):
        """Command to set ID target phase for movement [mm]."""
        return self.kparameter_set(phase, timeout)

    def phase_speed_set(self, phase_speed, timeout=None):
        """Command to set ID cruise phase speed for movement [mm/s]."""
        return self.kparameter_speed_set(phase_speed, timeout)

    def phase_speed_max_set(self, phase_speed_max, timeout=None):
        """Command to set ID max cruise phase speed for movement [mm/s]."""
        return self._write_sp('MaxPhaseSpeed-SP', phase_speed_max, timeout)

    # -- cmd_move

    def cmd_move_stop(self, timeout=None):
        """Send command to stop ID movement."""
        self['DevCtrl-Cmd'] = self._CMD_MOVE_STOP
        return True

    # --- private methods ---

    def _write_sp(self, propties_sp, values, timeout=None):
        pvs_sp_rb = ('Phase-SP', 'PhaseSpeed-SP')
        return super()._write_sp(
            propties_sp, values, timeout=timeout, pvs_sp_rb=pvs_sp_rb)

    def _move_start(
            self, cmd_propty, timeout=None, cmd_value=_CMD_MOVE_START):
        return super()._move_start(cmd_propty, timeout, cmd_value)


class PAPU(_ID):
    """PAPU Insertion Device."""

    class DEVICES:
        """Device names."""

        PAPU50_17SA = 'SI-17SA:ID-PAPU50'
        ALL = (PAPU50_17SA, )

    # --- PARAM_PVS ---
    PARAM_PVS = _PARAM_PVS()
    PARAM_PVS.PERIOD_LEN_CTE = 'PeriodLength-Cte'
    PARAM_PVS.KPARAM_SP = 'Phase-SP'
    PARAM_PVS.KPARAM_RB = 'Phase-RB'
    PARAM_PVS.KPARAM_MON = 'Phase-Mon'
    PARAM_PVS.KPARAM_PARKED_CTE = 'ParkedPhase-Cte'
    PARAM_PVS.KPARAM_MAXVELO_SP = 'MaxPhaseSpeed-SP'
    PARAM_PVS.KPARAM_MAXVELO_RB = 'MaxPhaseSpeed-RB'
    PARAM_PVS.KPARAM_VELO_SP = 'PhaseSpeed-SP'
    PARAM_PVS.KPARAM_VELO_RB = 'PhaseSpeed-RB'
    PARAM_PVS.KPARAM_VELO_MON = 'PhaseSpeed-Mon'
    PARAM_PVS.KPARAM_CHANGE_CMD = 'ChangePhase-Cmd'

    _properties = (
        'PwrPhase-Mon',
        'EnblAndReleasePhase-Sel', 'EnblAndReleasePhase-Sts',
        'AllowedToChangePhase-Mon',
        'StopPhase-Cmd', 'Log-Mon',
        )
    _properties_papu = (
        'Home-Cmd', 'EnblPwrPhase-Cmd', 'ClearErr-Cmd',
        'BeamLineCtrl-Mon',
        )

    PROPERTIES_DEFAULT = \
        tuple(set(value for key, value in _inspect.getmembers(PARAM_PVS) \
            if not key.startswith('_') and value is not None)) + \
        _properties + _properties_papu

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname is None:
            devname = self.DEVICES.PAPU50_17SA
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)

    @property
    def log_mon(self):
        """Return ID Log."""
        return self['Log-Mon']

    # --- phase speeds ----

    @property
    def phase_speed_max(self):
        """Return max phase speed readback [mm/s]."""
        return self.kparameter_speed_max

    @property
    def phase_speed_max_lims(self):
        """Return max phase speed limits."""
        return self.kparameter_speed_max_lims

    @property
    def phase_speed(self):
        """Return phase speed readback [mm/s]."""
        return self.kparameter_speed

    @property
    def phase_speed_mon(self):
        """Return phase speed monitor [mm/s]."""
        return self.kparameter_speed_mon

    # --- phase ---

    @property
    def phase_parked(self):
        """Return ID parked phase value [mm]."""
        return self.kparameter_parked

    @property
    def phase_lims(self):
        """Return ID phase limits [mm]."""
        return self.kparameter_lims

    @property
    def phase(self):
        """Return ID phase readback [mm]."""
        return self.kparameter

    @property
    def phase_mon(self):
        """Return ID phase monitor [mm]."""
        return self.kparameter_mon

    # --- drive checks ---

    @property
    def is_phase_drive_powered(self):
        """Return phase driver power state on (True|False)."""
        return self['PwrPhase-Mon'] != 0

    # --- movement checks ---

    @property
    def is_move_phase_enabled(self):
        """Return phase movement enabled state (True|False)."""
        return self['AllowedToChangePhase-Mon'] != 0

    @property
    def is_moving(self):
        """Return is moving state (True|False)."""
        return self[self.PARAM_PVS.IS_MOVING] != 0

    @property
    def is_homing(self):
        """Return whether ID is in homing procedure (True|False)."""
        return self['Home-Mon'] != 0

    # --- cmd_drive

    def cmd_drive_turn_power_on(self, timeout=None):
        """Command turn phase drive on."""
        if self.is_phase_drive_powered:
            return True
        self['EnblPwrPhase-Cmd'] = 1
        props_values = {'PwrPhase-Mon': 1}
        return self._wait_propty_values(props_values, timeout=timeout)

    # --- set methods ---

    def phase_set(self, phase, timeout=None):
        """Command to set ID target phase for movement [mm]."""
        return self.kparameter_set(phase, timeout)

    def phase_speed_set(self, phase_speed, timeout=None):
        """Command to set ID cruise phase speed for movement [mm/s]."""
        return self.kparameter_speed_set(phase_speed, timeout)

    def phase_speed_max_set(self, phase_speed_max, timeout=None):
        """Command to set ID max cruise phase speed for movement [mm/s]."""
        return self.kparameter_speed_max_set(phase_speed_max, timeout)

    # --- cmd_move disable/enable ---

    def cmd_move_phase_enable(self, timeout=None):
        """Command to release and enable ID phase movement."""
        self['EnblAndReleasePhase-Sel'] = 1
        return super()._wait(
            'AllowedToChangePhase-Mon',
            1, timeout=timeout, comp='eq')

    def cmd_move_phase_disable(self, timeout=None):
        """Command to disable and break ID phase movement."""
        self['EnblAndReleasePhase-Sel'] = 0
        return super()._wait(
            'AllowedToChangePhase-Mon',
            0, timeout=timeout, comp='eq')

    def cmd_move_enable(self, timeout=None):
        """Command to release and enable ID phase and gap movements."""
        return self.cmd_move_phase_enable(timeout=timeout)

    def cmd_move_disable(self, timeout=None):
        """Command to disable and break ID phase and gap movements."""
        return self.cmd_move_phase_disable(timeout=timeout)

    # -- cmd_move

    def cmd_move_phase_start(self, timeout=None):
        """Command to start phase movement."""
        return self.cmd_move_kparameter_start(timeout=timeout)

    # def cmd_move_park(self, timeout=None):
    #     """Command to set and start ID movement to parked config."""
    #     return super().cmd_move_park(timeout)

    # --- cmd_reset

    def cmd_device_reset(self, timeout=None):
        """Command to reset ID to a standard movement state."""
        success = True
        success &= self.cmd_beamline_ctrl_disable(timeout=timeout)
        success &= self.cmd_drive_turn_power_on(timeout=timeout)
        success &= self.cmd_move_enable(timeout=timeout)
        return success

    # --- other cmds ---

    def cmd_clear_error(self):
        """Command to clear errors."""
        self['ClearErr-Cmd'] = 1


class EPU(PAPU):
    """EPU Insertion Device."""

    class DEVICES:
        """Device names."""

        EPU50_10SB = 'SI-10SB:ID-EPU50'
        ALL = (EPU50_10SB, )

    # --- PARAM_PVS ---
    PARAM_PVS = _PARAM_PVS()
    PARAM_PVS.PERIOD_LEN_CTE = 'PeriodLength-Cte'
    PARAM_PVS.IS_MOVING = 'IsMoving-Mon'
    PARAM_PVS.PPARAM_SP = 'Phase-SP'
    PARAM_PVS.PPARAM_RB = 'Phase-RB'
    PARAM_PVS.PPARAM_MON = 'Phase-Mon'
    PARAM_PVS.PPARAM_PARKED_CTE = 'ParkedPhase-Cte'
    PARAM_PVS.PPARAM_MAXVELO_SP = 'MaxPhaseSpeed-SP'
    PARAM_PVS.PPARAM_MAXVELO_RB = 'MaxPhaseSpeed-RB'
    PARAM_PVS.PPARAM_VELO_SP = 'PhaseSpeed-SP'
    PARAM_PVS.PPARAM_VELO_RB = 'PhaseSpeed-RB'
    PARAM_PVS.PPARAM_VELO_MON = 'PhaseSpeed-Mon'
    PARAM_PVS.PPARAM_CHANGE_CMD = 'ChangePhase-Cmd'
    PARAM_PVS.KPARAM_SP = 'Gap-SP'
    PARAM_PVS.KPARAM_RB = 'Gap-RB'
    PARAM_PVS.KPARAM_MON = 'Gap-Mon'
    PARAM_PVS.KPARAM_PARKED_CTE = 'ParkedGap-Cte'
    PARAM_PVS.KPARAM_MAXVELO_SP = 'MaxGapSpeed-SP'
    PARAM_PVS.KPARAM_MAXVELO_RB = 'MaxGapSpeed-RB'
    PARAM_PVS.KPARAM_VELO_SP = 'GapSpeed-SP'
    PARAM_PVS.KPARAM_VELO_RB = 'GapSpeed-RB'
    PARAM_PVS.KPARAM_VELO_MON = 'GapSpeed-Mon'
    PARAM_PVS.KPARAM_CHANGE_CMD = 'ChangeGap-Cmd'
    PARAM_PVS.POL_SEL = 'Polarization-Sel'
    PARAM_PVS.POL_STS = 'Polarization-Sts'
    PARAM_PVS.POL_MON = 'Polarization-Mon'
    PARAM_PVS.POL_CHANGE_CMD = 'ChangePolarization-Cmd'

    PROPERTIES_DEFAULT = PAPU._properties + \
        tuple(set(value for key, value in _inspect.getmembers(PARAM_PVS) \
            if not key.startswith('_') and value is not None)) + (
        'EnblPwrAll-Cmd', 'PwrGap-Mon', 'Status-Mon',
        'EnblAndReleaseGap-Sel', 'EnblAndReleaseGap-Sts',
        'AllowedToChangeGap-Mon',
        'IsBusy-Mon', 'Stop-Cmd',
        )

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname is None:
            devname = self.DEVICES.EPU50_10SB
        if devname not in EPU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)

    @property
    def status(self):
        """ID status."""
        return self['Status-Mon']

    # --- gap speeds ----

    @property
    def gap_speed(self):
        """Return gap speed readback [mm/s]."""
        return self.kparameter_speed

    @property
    def gap_speed_mon(self):
        """Return gap speed monitor [mm/s]."""
        return self.kparameter_speed_mon

    @property
    def gap_speed_max(self):
        """Return max gap speed readback [mm/s]."""
        return self.kparameter_speed_max

    @property
    def gap_speed_max_lims(self):
        """Return max gap speed limits."""
        return self.kparameter_speed_max_lims

    # --- gap ---

    @property
    def gap_parked(self):
        """Return ID parked gap value [mm]."""
        return self.kparameter_parked

    @property
    def gap(self):
        """Return ID gap readback [mm]."""
        return self.kparameter

    @property
    def gap_lims(self):
        """Return ID gap control limits [mm]."""
        return self.kparameter_lims

    @property
    def gap_mon(self):
        """Return ID gap monitor [mm]."""
        return self.kparameter_mon

    # --- drive checks ---

    @property
    def is_gap_drive_powered(self):
        """Return gap driver power state on (True|False)."""
        return self['PwrGap-Mon'] != 0

    @property
    def is_drives_powered(self):
        """Return phase & gap drives powered state on (True|False)."""
        return self.is_phase_drive_powered and self.is_gap_drive_powered

    # --- movement checks ---

    @property
    def is_move_gap_enabled(self):
        """Return gap movement enabled state (True|False)."""
        return self['AllowedToChangeGap-Mon'] != 0

    @property
    def is_move_enabled(self):
        """Return phase and gap movements enabled state (True|False)."""
        return self.is_move_phase_enabled and self.is_move_gap_enabled

    @property
    def is_homing(self):
        """Return whether ID is in homing procedure."""
        return False

    # --- other checks ---

    @property
    def is_busy(self):
        """Return is busy state (True|False)."""
        return self['IsBusy-Mon'] != 0

    # --- cmd_beamline and cmd_drive

    def cmd_drive_turn_power_on(self, timeout=None):
        """Command turn phase and gap drives on."""
        if self.is_phase_drive_powered and self.is_gap_drive_powered:
            return True
        self['EnblPwrAll-Cmd'] = 1
        props_values = {'PwrPhase-Mon': 1, 'PwrGap-Mon': 1}
        return self._wait_set(props_values, timeout=timeout)

    # --- set methods ---

    def gap_set(self, gap, timeout=None):
        """Set ID target gap for movement [mm]."""
        return self.kparameter_set(gap, timeout)

    def gap_speed_set(self, gap_speed, timeout=None):
        """Set ID cruise gap speed for movement [mm/s]."""
        return self.kparameter_speed_set(gap_speed, timeout)

    def gap_speed_max_set(self, gap_speed_max, timeout=None):
        """Set ID max cruise gap speed for movement [mm/s]."""
        return self.kparameter_speed_max_set(gap_speed_max, timeout)

    # --- cmd_move disable/enable ---

    def cmd_move_gap_enable(self, timeout=None):
        """Command to release and enable ID gap movement."""
        self['EnblAndReleaseGap-Sel'] = 1
        return super()._wait(
            'AllowedToChangeGap-Mon',
            1, timeout=timeout, comp='eq')

    def cmd_move_gap_disable(self, timeout=None):
        """Command to disable and break ID gap movement."""
        self['EnblAndReleaseGap-Sel'] = 0
        return super()._wait(
            'AllowedToChangeGap-Mon',
            0, timeout=timeout, comp='eq')

    def cmd_move_enable(self, timeout=None):
        """Command to release and enable ID phase and gap movements."""
        success = True
        success &= self.cmd_move_phase_enable(timeout=timeout)
        success &= self.cmd_move_gap_enable(timeout=timeout)
        return success

    def cmd_move_disable(self, timeout=None):
        """Command to disable and break ID phase and gap movements."""
        success = True
        success &= self.cmd_move_phase_disable(timeout=timeout)
        success &= self.cmd_move_gap_disable(timeout=timeout)
        return success

    # --- cmd_wait

    def wait_while_busy(self, timeout=None):
        """Command wait within timeout while ID control is busy."""
        timeout = timeout or self._DEF_TIMEOUT
        time_init = _time.time()
        while self.is_busy:
            _time.sleep(min(self._SHORT_SHUT_EYE, timeout))
            if _time.time() - time_init > timeout:
                return False
        return True

    # -- cmd_move

    def cmd_move_gap_start(self, timeout=None):
        """Command to start gap movement."""
        return self.cmd_move_kparameter_start(timeout)

    # --- other cmds ---

    def cmd_clear_error(self):
        """Command to clear errors."""
        pass

    # --- private methods ---

    @staticmethod
    def _calc_eta_select_time(dtime_kparam, dtime_pparam):
        return max(dtime_kparam, dtime_pparam)


class DELTA(_ID):
    """DELTA Insertion Device."""

    class DEVICES:
        """Device names."""

        DELTA52_10SB = 'SI-10SB:ID-DELTA52'
        ALL = (DELTA52_10SB, )

    # --- PARAM_PVS ---
    PARAM_PVS = _PARAM_PVS()
    PARAM_PVS.PERIOD_LEN_CTE = 'PeriodLength-Cte'
    PARAM_PVS.IS_MOVING = 'Moving-Mon'
    PARAM_PVS.START_PARKING_CMD = 'StartParking-Cmd'
    PARAM_PVS.MOVE_ABORT = 'Abort-Cmd'
    PARAM_PVS.PPARAM_SP = 'PParam-SP'
    PARAM_PVS.PPARAM_RB = 'PParam-RB'
    PARAM_PVS.PPARAM_MON = 'PParam-Mon'
    PARAM_PVS.PPARAM_PARKED_CTE = 'PParamParked-Cte'
    PARAM_PVS.PPARAM_MAXVELO_SP = 'MaxVelo-SP'
    PARAM_PVS.PPARAM_MAXVELO_RB = 'MaxVelo-RB'
    PARAM_PVS.PPARAM_VELO_SP = 'PParamVelo-SP'
    PARAM_PVS.PPARAM_VELO_RB = 'PParamVelo-RB'
    PARAM_PVS.PPARAM_TOL_SP = 'PolTol-SP'
    PARAM_PVS.PPARAM_TOL_RB = 'PolTol-RB'
    PARAM_PVS.PPARAM_CHANGE_CMD = 'PParamChange-Cmd'
    PARAM_PVS.KPARAM_SP = 'KParam-SP'
    PARAM_PVS.KPARAM_RB = 'KParam-RB'
    PARAM_PVS.KPARAM_MON = 'KParam-Mon'
    PARAM_PVS.KPARAM_PARKED_CTE = 'KParamParked-Cte'
    PARAM_PVS.KPARAM_MAXVELO_SP = 'MaxVelo-SP'
    PARAM_PVS.KPARAM_MAXVELO_RB = 'MaxVelo-RB'
    PARAM_PVS.KPARAM_VELO_SP = 'KParamVelo-SP'
    PARAM_PVS.KPARAM_VELO_RB = 'KParamVelo-RB'
    PARAM_PVS.KPARAM_TOL_SP = 'PosTol-SP'
    PARAM_PVS.KPARAM_TOL_RB = 'PosTol-RB'
    PARAM_PVS.KPARAM_CHANGE_CMD = 'KParamChange-Cmd'
    PARAM_PVS.POL_SEL = 'Pol-Sel'
    PARAM_PVS.POL_STS = 'Pol-Sts'
    PARAM_PVS.POL_MON = 'Pol-Mon'
    PARAM_PVS.POL_CHANGE_CMD = 'PolChange-Cmd'

    PROPERTIES_DEFAULT = tuple(set(
        value for key, value in _inspect.getmembers(PARAM_PVS)
        if not key.startswith('_') and value is not None))
    PROPERTIES_DEFAULT = PROPERTIES_DEFAULT + (
        'CSDVirtPos-Mon', 'CSEVirtPos-Mon',
        'CIEVirtPos-Mon', 'CIDVirtPos-Mon',
        'IsOperational-Mon', 'MotorsEnbld-Mon',
        'Alarm-Mon', 'Intlk-Mon', 'IntlkBits-Mon', 'IntlkLabels-Cte',
        'ConsistentSetPoints-Mon', 'PLCState-Mon',
        'Energy-SP', 'Energy-RB', 'Energy-Mon',
        'KValue-SP', 'KValue-RB', 'KValue-Mon',
        )

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname is None:
            devname = self.DEVICES.DELTA52_10SB
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)

    @property
    def is_operational(self):
        """Return True if ID is operational."""
        return self['IsOperational-Mon'] == 0  # 0 : 'OK'

    # --- cassette positions ---

    @property
    def pos_csd_mon(self):
        """Return longitudinal position of CSD [mm].

        cassette positions x (PParam, KParam):
            pos_cid = PParam
            pos_cse = PParam + KParam
            pos_csd = KParam
            pos_cie = 0
        """
        return self['CSDVirtPos-Mon']

    @property
    def pos_cse_mon(self):
        """Return longitudinal position of CSE [mm].

        cassette positions x (PParam, KParam):
            pos_cid = PParam
            pos_cse = PParam + KParam
            pos_csd = KParam
            pos_cie = 0
        """
        return self['CSEVirtPos-Mon']

    @property
    def pos_cie_mon(self):
        """Return longitudinal position of CIE [mm].

        cassette positions x (PParam, KParam):
            pos_cid = PParam
            pos_cse = PParam + KParam
            pos_csd = KParam
            pos_cie = 0
        """
        return self['CIEVirtPos-Mon']

    @property
    def pos_cid_mon(self):
        """Return longitudinal position of CID [mm].

        cassette positions x (PParam, KParam):
            pos_cid = PParam
            pos_cse = PParam + KParam
            pos_csd = KParam
            pos_cie = 0
        """
        return self['CIDVirtPos-Mon']

    def _calc_eta(self, pparam, kparam):
        """."""
        if kparam is not None and self.polarization_mon_str == 'circularp':
            kparam = -kparam
        return super()._calc_eta(pparam, kparam)


class WIG(_ID):
    """Wiggler Insertion Device."""

    class DEVICES:
        """Device names."""

        WIG180_14SB = 'SI-14SB:ID-WIG180'
        ALL = (WIG180_14SB, )

    # NOTE: IOC yet to be written!
    PROPERTIES_DEFAULT = ('Gap-SP', 'Gap-RB', 'Gap-Mon')

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname is None:
            devname = self.DEVICES.WIG180_14SB
        if devname not in WIG.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)
